# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# service type constants:
CORE = "CORE"
DUMMY = "DUMMY"
LOADBALANCER = "LOADBALANCER"
FIREWALL = "FIREWALL"
VPN = "VPN"
METERING = "METERING"
L3_ROUTER_NAT = "L3_ROUTER_NAT"


#maps extension alias to service type
EXT_TO_SERVICE_MAPPING = {
    'dummy': DUMMY,
    'lbaas': LOADBALANCER,
    'fwaas': FIREWALL,
    'vpnaas': VPN,
    'metering': METERING,
    'router': L3_ROUTER_NAT
}

# TODO(salvatore-orlando): Move these (or derive them) from conf file
ALLOWED_SERVICES = [CORE, DUMMY, LOADBALANCER, FIREWALL, VPN, METERING,
                    L3_ROUTER_NAT]

COMMON_PREFIXES = {
    CORE: "",
    DUMMY: "/dummy_svc",
    LOADBALANCER: "/lb",
    FIREWALL: "/fw",
    VPN: "/vpn",
    METERING: "/metering",
    L3_ROUTER_NAT: "",
}

# Service operation status constants
ACTIVE = "ACTIVE"
DOWN = "DOWN"
PENDING_CREATE = "PENDING_CREATE"
PENDING_UPDATE = "PENDING_UPDATE"
PENDING_DELETE = "PENDING_DELETE"
INACTIVE = "INACTIVE"
ERROR = "ERROR"

ACTIVE_PENDING_STATUSES = (
    ACTIVE,
    PENDING_CREATE,
    PENDING_UPDATE
)

# FWaaS firewall rule action
FWAAS_ALLOW = "allow"
FWAAS_DENY = "deny"

# L3 Protocol name constants
TCP = "tcp"
UDP = "udp"
ICMP = "icmp"

# Network Type constants
TYPE_FLAT = 'flat'
TYPE_GRE = 'gre'
TYPE_LOCAL = 'local'
TYPE_VXLAN = 'vxlan'
TYPE_VLAN = 'vlan'
TYPE_NONE = 'none'



# Copyright 2014 OneConvergence, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Kedar Kulkarni, One Convergence, Inc.

import contextlib
import time

import mock
from oslo.config import cfg
import testtools

from neutron.agent.linux import ovs_lib
from neutron.extensions import securitygroup as ext_sg
from neutron.plugins.oneconvergence.agent import nvsd_neutron_agent
from neutron.tests import base

DAEMON_LOOP_COUNT = 5


class TestOneConvergenceAgentBase(base.BaseTestCase):

    def setUp(self):
        super(TestOneConvergenceAgentBase, self).setUp()
        cfg.CONF.set_default('firewall_driver',
                             'neutron.agent.firewall.NoopFirewallDriver',
                             group='SECURITYGROUP')
        cfg.CONF.set_override('rpc_backend',
                              'neutron.openstack.common.rpc.impl_fake')
        with contextlib.nested(
            mock.patch('neutron.openstack.common.loopingcall.'
                       'FixedIntervalLoopingCall'),
        ) as (loopingcall):
            kwargs = {'integ_br': 'integration_bridge',
                      'root_helper': 'dummy_wrapper',
                      'polling_interval': 5}
            context = mock.Mock()
            self.agent = nvsd_neutron_agent.NVSDNeutronAgent(**kwargs)
            self.sg_agent = nvsd_neutron_agent.SecurityGroupAgentRpc(
                context, 'dummy_wrapper')
            self.callback_nvsd = nvsd_neutron_agent.NVSDAgentRpcCallback(
                context, self.agent, self.sg_agent)
            self.loopingcall = loopingcall


class TestOneConvergenceAgentCallback(TestOneConvergenceAgentBase):

    def test_port_update(self):
        with contextlib.nested(
            mock.patch.object(ovs_lib.OVSBridge, 'get_vif_port_by_id'),
            mock.patch.object(self.sg_agent, 'refresh_firewall')
        ) as (get_vif_port_by_id, refresh_firewall):
            context = mock.Mock()
            vifport = ovs_lib.VifPort('port1', '1', 'id-1', 'mac-1',
                                      self.agent.int_br)

            # The OVS port does not exist.
            get_vif_port_by_id.return_value = None
            port = {'id': 'update-port-1'}
            self.callback_nvsd.port_update(context, port=port)
            self.assertEqual(get_vif_port_by_id.call_count, 1)
            self.assertFalse(refresh_firewall.call_count)

            # The OVS port exists but no security group is associated.
            get_vif_port_by_id.return_value = vifport
            port = {'id': 'update-port-1'}
            self.callback_nvsd.port_update(context, port=port)
            self.assertEqual(get_vif_port_by_id.call_count, 2)
            self.assertFalse(refresh_firewall.call_count)

            # The OVS port exists but a security group is associated.
            get_vif_port_by_id.return_value = vifport
            port = {'id': 'update-port-1',
                    ext_sg.SECURITYGROUPS: ['default']}
            self.callback_nvsd.port_update(context, port=port)
            self.assertEqual(get_vif_port_by_id.call_count, 3)
            self.assertEqual(refresh_firewall.call_count, 1)

            get_vif_port_by_id.return_value = None
            port = {'id': 'update-port-1',
                    ext_sg.SECURITYGROUPS: ['default']}
            self.callback_nvsd.port_update(context, port=port)
            self.assertEqual(get_vif_port_by_id.call_count, 4)
            self.assertEqual(refresh_firewall.call_count, 1)


class TestNVSDAgent(TestOneConvergenceAgentBase):

    def _setup_mock(self):
        self.get_vif_ports = mock.patch.object(
            ovs_lib.OVSBridge, 'get_vif_port_set',
            return_value=set(['id-1', 'id-2'])).start()
        self.prepare_devices_filter = mock.patch.object(
            self.agent.sg_agent, 'prepare_devices_filter').start()
        self.remove_devices_filter = mock.patch.object(
            self.agent.sg_agent, 'remove_devices_filter').start()

    def test_daemon_loop(self):

        def state_check(index):
            self.assertEqual(len(self.vif_ports_scenario[index]),
                             len(self.agent.ports))

        # Fake time.sleep to stop the infinite loop in daemon_loop()
        self.sleep_count = 0

        def sleep_mock(*args, **kwargs):
            state_check(self.sleep_count)
            self.sleep_count += 1
            if self.sleep_count >= DAEMON_LOOP_COUNT:
                raise RuntimeError()

        self.vif_ports_scenario = [set(), set(), set(), set(['id-1', 'id-2']),
                                   set(['id-2', 'id-3'])]

        # Ensure vif_ports_scenario is longer than DAEMON_LOOP_COUNT
        if len(self.vif_ports_scenario) < DAEMON_LOOP_COUNT:
            self.vif_ports_scenario.extend(
                [] for _i in xrange(DAEMON_LOOP_COUNT -
                                    len(self.vif_ports_scenario)))

        with contextlib.nested(
            mock.patch.object(time, 'sleep', side_effect=sleep_mock),
            mock.patch.object(ovs_lib.OVSBridge, 'get_vif_port_set'),
            mock.patch.object(self.agent.sg_agent, 'prepare_devices_filter'),
            mock.patch.object(self.agent.sg_agent, 'remove_devices_filter')
        ) as (sleep, get_vif_port_set, prepare_devices_filter,
              remove_devices_filter):
            get_vif_port_set.side_effect = self.vif_ports_scenario

            with testtools.ExpectedException(RuntimeError):
                self.agent.daemon_loop()
            self.assertEqual(sleep.call_count, DAEMON_LOOP_COUNT)

            expected = [mock.call(set(['id-1', 'id-2'])),
                        mock.call(set(['id-3']))]

            self.assertEqual(prepare_devices_filter.call_count, 2)
            prepare_devices_filter.assert_has_calls(expected)

            expected = [mock.call(set([])), mock.call(set(['id-1']))]

            self.assertEqual(remove_devices_filter.call_count, 2)
            remove_devices_filter.assert_has_calls(expected)

            sleep.assert_called_with(self.agent.polling_interval)


class TestOneConvergenceAgentMain(base.BaseTestCase):
    def test_main(self):
        with contextlib.nested(
            mock.patch.object(nvsd_neutron_agent, 'NVSDNeutronAgent'),
            mock.patch('eventlet.monkey_patch'),
            mock.patch.object(nvsd_neutron_agent, 'logging_config'),
            mock.patch.object(nvsd_neutron_agent, 'config')
        ) as (agent, eventlet, logging_config, config):
            config.AGENT.integration_bridge = 'br-int-dummy'
            config.AGENT.root_helper = 'root-helper'
            config.AGENT.polling_interval = 5

            nvsd_neutron_agent.main()

            self.assertTrue(eventlet.called)
            self.assertTrue(logging_config.setup_logging.called)
            agent.assert_has_calls([
                mock.call('br-int-dummy', 'root-helper', 5),
                mock.call().daemon_loop()
            ])



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Mohammad Banikazemi, IBM Corp.


import httplib
import urllib

import httplib2
from keystoneclient.v2_0 import client as keyclient
from oslo.config import cfg

from neutron.api.v2 import attributes
from neutron.openstack.common import log as logging
from neutron.plugins.ibm.common import config  # noqa
from neutron.plugins.ibm.common import constants
from neutron.wsgi import Serializer

LOG = logging.getLogger(__name__)

SDNVE_VERSION = '2.0'
SDNVE_ACTION_PREFIX = '/sdnve'
SDNVE_RETRIES = 0
SDNVE_RETRIY_INTERVAL = 1
SDNVE_TENANT_TYPE_OVERLAY = u'DOVE'
SDNVE_URL = 'https://%s:%s%s'


class RequestHandler(object):
    '''Handles processeing requests to and responses from controller.'''

    def __init__(self, controller_ips=None, port=None, ssl=None,
                 base_url=None, userid=None, password=None,
                 timeout=10, formats=None):
        '''Initializes the RequestHandler for communication with controller

        Following keyword arguments are used; if not specified, default
        values are used.
        :param port: Username for authentication.
        :param timeout: Time out for http requests.
        :param userid: User id for accessing controller.
        :param password: Password for accessing the controlelr.
        :param base_url: The base url for the controller.
        :param controller_ips: List of controller IP addresses.
        :param formats: Supported formats.
        '''
        self.port = port or cfg.CONF.SDNVE.port
        self.timeout = timeout
        self._s_meta = None
        self.connection = None
        self.httpclient = httplib2.Http(
            disable_ssl_certificate_validation=True)
        self.cookie = None

        userid = userid or cfg.CONF.SDNVE.userid
        password = password or cfg.CONF.SDNVE.password
        if (userid and password):
            self.httpclient.add_credentials(userid, password)

        self.base_url = base_url or cfg.CONF.SDNVE.base_url
        self.controller_ips = controller_ips or cfg.CONF.SDNVE.controller_ips

        LOG.info(_("The IP addr of available SDN-VE controllers: %s"),
                 self.controller_ips)
        self.controller_ip = self.controller_ips[0]
        LOG.info(_("The SDN-VE controller IP address: %s"),
                 self.controller_ip)

        self.new_controller = False
        self.format = formats or cfg.CONF.SDNVE.format

        self.version = SDNVE_VERSION
        self.action_prefix = SDNVE_ACTION_PREFIX
        self.retries = SDNVE_RETRIES
        self.retry_interval = SDNVE_RETRIY_INTERVAL

    def serialize(self, data):
        '''Serializes a dictionary with a single key.'''

        if isinstance(data, dict):
            return Serializer().serialize(data, self.content_type())
        elif data:
            raise TypeError(_("unable to serialize object type: '%s'") %
                            type(data))

    def deserialize(self, data, status_code):
        '''Deserializes an xml or json string into a dictionary.'''

        # NOTE(mb): Temporary fix for backend controller requirement
        data = data.replace("router_external", "router:external")

        if status_code == httplib.NO_CONTENT:
            return data
        try:
            deserialized_data = Serializer(
                metadata=self._s_meta).deserialize(data, self.content_type())
            deserialized_data = deserialized_data['body']
        except Exception:
            deserialized_data = data

        return deserialized_data

    def content_type(self, format=None):
        '''Returns the mime-type for either 'xml' or 'json'.'''

        return 'application/%s' % (format or self.format)

    def delete(self, url, body=None, headers=None, params=None):
        return self.do_request("DELETE", url, body=body,
                               headers=headers, params=params)

    def get(self, url, body=None, headers=None, params=None):
        return self.do_request("GET", url, body=body,
                               headers=headers, params=params)

    def post(self, url, body=None, headers=None, params=None):
        return self.do_request("POST", url, body=body,
                               headers=headers, params=params)

    def put(self, url, body=None, headers=None, params=None):
        return self.do_request("PUT", url, body=body,
                               headers=headers, params=params)

    def do_request(self, method, url, body=None, headers=None,
                   params=None, connection_type=None):

        status_code = -1
        replybody_deserialized = ''

        if body:
            body = self.serialize(body)

        self.headers = headers or {'Content-Type': self.content_type()}
        if self.cookie:
            self.headers['cookie'] = self.cookie

        if self.controller_ip != self.controller_ips[0]:
            controllers = [self.controller_ip]
        else:
            controllers = []
        controllers.extend(self.controller_ips)

        for controller_ip in controllers:
            serverurl = SDNVE_URL % (controller_ip, self.port, self.base_url)
            myurl = serverurl + url
            if params and isinstance(params, dict):
                myurl += '?' + urllib.urlencode(params, doseq=1)

            try:
                LOG.debug(_("Sending request to SDN-VE. url: "
                            "%(myurl)s method: %(method)s body: "
                            "%(body)s header: %(header)s "),
                          {'myurl': myurl, 'method': method,
                           'body': body, 'header': self.headers})
                resp, replybody = self.httpclient.request(
                    myurl, method=method, body=body, headers=self.headers)
                LOG.debug(("Response recd from SDN-VE. resp: %(resp)s"
                           "body: %(body)s"),
                          {'resp': resp.status, 'body': replybody})
                status_code = resp.status

            except Exception as e:
                LOG.error(_("Error: Could not reach server: %(url)s "
                            "Exception: %(excp)s."),
                          {'url': myurl, 'excp': e})
                self.cookie = None
                continue

            if status_code not in constants.HTTP_ACCEPTABLE:
                LOG.debug(_("Error message: %(reply)s --  Status: %(status)s"),
                          {'reply': replybody, 'status': status_code})
            else:
                LOG.debug(_("Received response status: %s"), status_code)

            if resp.get('set-cookie'):
                self.cookie = resp['set-cookie']
            replybody_deserialized = self.deserialize(
                replybody,
                status_code)
            LOG.debug(_("Deserialized body: %s"), replybody_deserialized)
            if controller_ip != self.controller_ip:
                # bcast the change of controller
                self.new_controller = True
                self.controller_ip = controller_ip

            return (status_code, replybody_deserialized)

        return (httplib.REQUEST_TIMEOUT, 'Could not reach server(s)')


class Client(RequestHandler):
    '''Client for SDNVE controller.'''

    def __init__(self):
        '''Initialize a new SDNVE client.'''
        super(Client, self).__init__()

        self.keystoneclient = KeystoneClient()

    resource_path = {
        'network': "ln/networks/",
        'subnet': "ln/subnets/",
        'port': "ln/ports/",
        'tenant': "ln/tenants/",
        'router': "ln/routers/",
        'floatingip': "ln/floatingips/",
    }

    def process_request(self, body):
        '''Processes requests according to requirements of controller.'''
        if self.format == 'json':
            body = dict(
                (k.replace(':', '_'), v) for k, v in body.items()
                if attributes.is_attr_set(v))

    def sdnve_list(self, resource, **params):
        '''Fetches a list of resources.'''

        res = self.resource_path.get(resource, None)
        if not res:
            LOG.info(_("Bad resource for forming a list request"))
            return 0, ''

        return self.get(res, params=params)

    def sdnve_show(self, resource, specific, **params):
        '''Fetches information of a certain resource.'''

        res = self.resource_path.get(resource, None)
        if not res:
            LOG.info(_("Bad resource for forming a show request"))
            return 0, ''

        return self.get(res + specific, params=params)

    def sdnve_create(self, resource, body):
        '''Creates a new resource.'''

        res = self.resource_path.get(resource, None)
        if not res:
            LOG.info(_("Bad resource for forming a create request"))
            return 0, ''

        self.process_request(body)
        status, data = self.post(res, body=body)
        return (status, data)

    def sdnve_update(self, resource, specific, body=None):
        '''Updates a resource.'''

        res = self.resource_path.get(resource, None)
        if not res:
            LOG.info(_("Bad resource for forming a update request"))
            return 0, ''

        self.process_request(body)
        return self.put(res + specific, body=body)

    def sdnve_delete(self, resource, specific):
        '''Deletes the specified resource.'''

        res = self.resource_path.get(resource, None)
        if not res:
            LOG.info(_("Bad resource for forming a delete request"))
            return 0, ''

        return self.delete(res + specific)

    def _tenant_id_conversion(self, osid):
        return osid

    def sdnve_get_tenant_byid(self, os_tenant_id):
        sdnve_tenant_id = self._tenant_id_conversion(os_tenant_id)
        resp, content = self.sdnve_show('tenant', sdnve_tenant_id)
        if resp in constants.HTTP_ACCEPTABLE:
            tenant_id = content.get('id')
            tenant_type = content.get('network_type')
            if tenant_type == SDNVE_TENANT_TYPE_OVERLAY:
                tenant_type = constants.TENANT_TYPE_OVERLAY
            return tenant_id, tenant_type
        return None, None

    def sdnve_check_and_create_tenant(self, os_tenant_id, network_type=None):

        if not os_tenant_id:
            return
        tenant_id, tenant_type = self.sdnve_get_tenant_byid(os_tenant_id)
        if tenant_id:
            if not network_type:
                return tenant_id
            if tenant_type != network_type:
                LOG.info(_("Non matching tenant and network types: "
                           "%(ttype)s %(ntype)s"),
                         {'ttype': tenant_type, 'ntype': network_type})
                return
            return tenant_id

        # Have to create a new tenant
        sdnve_tenant_id = self._tenant_id_conversion(os_tenant_id)
        if not network_type:
            network_type = self.keystoneclient.get_tenant_type(os_tenant_id)
        if network_type == constants.TENANT_TYPE_OVERLAY:
            network_type = SDNVE_TENANT_TYPE_OVERLAY

        pinn_desc = ("Created by SDN-VE Neutron Plugin, OS project name = " +
                     self.keystoneclient.get_tenant_name(os_tenant_id))

        res, content = self.sdnve_create('tenant',
                                         {'id': sdnve_tenant_id,
                                          'name': os_tenant_id,
                                          'network_type': network_type,
                                          'description': pinn_desc})
        if res not in constants.HTTP_ACCEPTABLE:
            return

        return sdnve_tenant_id

    def sdnve_get_controller(self):
        if self.new_controller:
            self.new_controller = False
            return self.controller_ip


class KeystoneClient(object):

    def __init__(self, username=None, tenant_name=None, password=None,
                 auth_url=None):

        keystone_conf = cfg.CONF.keystone_authtoken
        keystone_auth_url = ('%s://%s:%s/v2.0/' %
                             (keystone_conf.auth_protocol,
                              keystone_conf.auth_host,
                              keystone_conf.auth_port))

        username = username or keystone_conf.admin_user
        tenant_name = tenant_name or keystone_conf.admin_tenant_name
        password = password or keystone_conf.admin_password
        auth_url = auth_url or keystone_auth_url

        self.overlay_signature = cfg.CONF.SDNVE.overlay_signature
        self.of_signature = cfg.CONF.SDNVE.of_signature
        self.default_tenant_type = cfg.CONF.SDNVE.default_tenant_type

        self.client = keyclient.Client(username=username,
                                       password=password,
                                       tenant_name=tenant_name,
                                       auth_url=auth_url)

    def get_tenant_byid(self, id):

        try:
            return self.client.tenants.get(id)
        except Exception:
            LOG.exception(_("Did not find tenant: %r"), id)

    def get_tenant_type(self, id):

        tenant = self.get_tenant_byid(id)
        if tenant:
            description = tenant.description
            if description:
                if (description.find(self.overlay_signature) >= 0):
                    return constants.TENANT_TYPE_OVERLAY
                if (description.find(self.of_signature) >= 0):
                    return constants.TENANT_TYPE_OF
        return self.default_tenant_type

    def get_tenant_name(self, id):

        tenant = self.get_tenant_byid(id)
        if tenant:
            return tenant.name
        return 'not found'



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @Author Don Kehn, dekehn@gmail.com

from neutron.api import extensions
from neutron.api.v2 import attributes as attr
from neutron.common import exceptions


# ExtraDHcpOpts Exceptions
class ExtraDhcpOptNotFound(exceptions.NotFound):
    message = _("ExtraDhcpOpt %(id)s could not be found")


class ExtraDhcpOptBadData(exceptions.InvalidInput):
    message = _("Invalid data format for extra-dhcp-opt: %(data)s")


def _validate_list_of_dict_or_none(data, key_specs=None):
    if data is not None:
        if not isinstance(data, list):
            raise ExtraDhcpOptBadData(data=data)
        for d in data:
            msg = attr._validate_dict(d, key_specs)
            if msg:
                raise ExtraDhcpOptBadData(data=msg)


attr.validators['type:list_of_dict_or_none'] = _validate_list_of_dict_or_none

# Attribute Map
EXTRADHCPOPTS = 'extra_dhcp_opts'

EXTENDED_ATTRIBUTES_2_0 = {
    'ports': {
        EXTRADHCPOPTS:
        {'allow_post': True,
         'allow_put': True,
         'is_visible': True,
         'default': None,
         'validate': {
             'type:list_of_dict_or_none': {
                 'id': {'type:uuid': None, 'required': False},
                 'opt_name': {'type:not_empty_string': None,
                              'required': True},
                 'opt_value': {'type:not_empty_string_or_none': None,
                               'required': True}}}}}}


class Extra_dhcp_opt(extensions.ExtensionDescriptor):
    @classmethod
    def get_name(cls):
        return "Neutron Extra DHCP opts"

    @classmethod
    def get_alias(cls):
        return "extra_dhcp_opt"

    @classmethod
    def get_description(cls):
        return ("Extra options configuration for DHCP. "
                "For example PXE boot options to DHCP clients can "
                "be specified (e.g. tftp-server, server-ip-address, "
                "bootfile-name)")

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/neutron/extra_dhcp_opt/api/v1.0"

    @classmethod
    def get_updated(cls):
        return "2013-03-17T12:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# @author: Kyle Mestery, Cisco Systems, Inc.

from oslo.config import cfg
import sqlalchemy as sa
from sqlalchemy.orm import exc as sa_exc

from neutron.common import exceptions as exc
from neutron.db import api as db_api
from neutron.db import model_base
from neutron.openstack.common import log
from neutron.plugins.common import constants as p_const
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import type_tunnel

LOG = log.getLogger(__name__)

VXLAN_UDP_PORT = 4789
MAX_VXLAN_VNI = 16777215

vxlan_opts = [
    cfg.ListOpt('vni_ranges',
                default=[],
                help=_("Comma-separated list of <vni_min>:<vni_max> tuples "
                       "enumerating ranges of VXLAN VNI IDs that are "
                       "available for tenant network allocation")),
    cfg.StrOpt('vxlan_group', default=None,
               help=_("Multicast group for VXLAN. If unset, disables VXLAN "
                      "multicast mode.")),
]

cfg.CONF.register_opts(vxlan_opts, "ml2_type_vxlan")


class VxlanAllocation(model_base.BASEV2):

    __tablename__ = 'ml2_vxlan_allocations'

    vxlan_vni = sa.Column(sa.Integer, nullable=False, primary_key=True,
                          autoincrement=False)
    allocated = sa.Column(sa.Boolean, nullable=False, default=False)


class VxlanEndpoints(model_base.BASEV2):
    """Represents tunnel endpoint in RPC mode."""
    __tablename__ = 'ml2_vxlan_endpoints'

    ip_address = sa.Column(sa.String(64), primary_key=True)
    udp_port = sa.Column(sa.Integer, primary_key=True, nullable=False,
                         autoincrement=False)

    def __repr__(self):
        return "<VxlanTunnelEndpoint(%s)>" % self.ip_address


class VxlanTypeDriver(type_tunnel.TunnelTypeDriver):

    def get_type(self):
        return p_const.TYPE_VXLAN

    def initialize(self):
        self.vxlan_vni_ranges = []
        self._parse_tunnel_ranges(
            cfg.CONF.ml2_type_vxlan.vni_ranges,
            self.vxlan_vni_ranges,
            p_const.TYPE_VXLAN
        )
        self._sync_vxlan_allocations()

    def reserve_provider_segment(self, session, segment):
        segmentation_id = segment.get(api.SEGMENTATION_ID)
        with session.begin(subtransactions=True):
            try:
                alloc = (session.query(VxlanAllocation).
                         filter_by(vxlan_vni=segmentation_id).
                         with_lockmode('update').
                         one())
                if alloc.allocated:
                    raise exc.TunnelIdInUse(tunnel_id=segmentation_id)
                LOG.debug(_("Reserving specific vxlan tunnel %s from pool"),
                          segmentation_id)
                alloc.allocated = True
            except sa_exc.NoResultFound:
                LOG.debug(_("Reserving specific vxlan tunnel %s outside pool"),
                          segmentation_id)
                alloc = VxlanAllocation(vxlan_vni=segmentation_id)
                alloc.allocated = True
                session.add(alloc)

    def allocate_tenant_segment(self, session):
        with session.begin(subtransactions=True):
            alloc = (session.query(VxlanAllocation).
                     filter_by(allocated=False).
                     with_lockmode('update').
                     first())
            if alloc:
                LOG.debug(_("Allocating vxlan tunnel vni %(vxlan_vni)s"),
                          {'vxlan_vni': alloc.vxlan_vni})
                alloc.allocated = True
                return {api.NETWORK_TYPE: p_const.TYPE_VXLAN,
                        api.PHYSICAL_NETWORK: None,
                        api.SEGMENTATION_ID: alloc.vxlan_vni}

    def release_segment(self, session, segment):
        vxlan_vni = segment[api.SEGMENTATION_ID]
        with session.begin(subtransactions=True):
            try:
                alloc = (session.query(VxlanAllocation).
                         filter_by(vxlan_vni=vxlan_vni).
                         with_lockmode('update').
                         one())
                alloc.allocated = False
                for low, high in self.vxlan_vni_ranges:
                    if low <= vxlan_vni <= high:
                        LOG.debug(_("Releasing vxlan tunnel %s to pool"),
                                  vxlan_vni)
                        break
                else:
                    session.delete(alloc)
                    LOG.debug(_("Releasing vxlan tunnel %s outside pool"),
                              vxlan_vni)
            except sa_exc.NoResultFound:
                LOG.warning(_("vxlan_vni %s not found"), vxlan_vni)

    def _sync_vxlan_allocations(self):
        """
        Synchronize vxlan_allocations table with configured tunnel ranges.
        """

        # determine current configured allocatable vnis
        vxlan_vnis = set()
        for tun_min, tun_max in self.vxlan_vni_ranges:
            if tun_max + 1 - tun_min > MAX_VXLAN_VNI:
                LOG.error(_("Skipping unreasonable VXLAN VNI range "
                            "%(tun_min)s:%(tun_max)s"),
                          {'tun_min': tun_min, 'tun_max': tun_max})
            else:
                vxlan_vnis |= set(xrange(tun_min, tun_max + 1))

        session = db_api.get_session()
        with session.begin(subtransactions=True):
            # remove from table unallocated tunnels not currently allocatable
            allocs = session.query(VxlanAllocation)
            for alloc in allocs:
                try:
                    # see if tunnel is allocatable
                    vxlan_vnis.remove(alloc.vxlan_vni)
                except KeyError:
                    # it's not allocatable, so check if its allocated
                    if not alloc.allocated:
                        # it's not, so remove it from table
                        LOG.debug(_("Removing tunnel %s from pool"),
                                  alloc.vxlan_vni)
                        session.delete(alloc)

            # add missing allocatable tunnels to table
            for vxlan_vni in sorted(vxlan_vnis):
                alloc = VxlanAllocation(vxlan_vni=vxlan_vni)
                session.add(alloc)

    def get_vxlan_allocation(self, session, vxlan_vni):
        with session.begin(subtransactions=True):
            return session.query(VxlanAllocation).filter_by(
                vxlan_vni=vxlan_vni).first()

    def get_endpoints(self):
        """Get every vxlan endpoints from database."""

        LOG.debug(_("get_vxlan_endpoints() called"))
        session = db_api.get_session()

        with session.begin(subtransactions=True):
            vxlan_endpoints = session.query(VxlanEndpoints)
            return [{'ip_address': vxlan_endpoint.ip_address,
                     'udp_port': vxlan_endpoint.udp_port}
                    for vxlan_endpoint in vxlan_endpoints]

    def add_endpoint(self, ip, udp_port=VXLAN_UDP_PORT):
        LOG.debug(_("add_vxlan_endpoint() called for ip %s"), ip)
        session = db_api.get_session()
        with session.begin(subtransactions=True):
            try:
                vxlan_endpoint = (session.query(VxlanEndpoints).
                                  filter_by(ip_address=ip).
                                  with_lockmode('update').one())
            except sa_exc.NoResultFound:
                vxlan_endpoint = VxlanEndpoints(ip_address=ip,
                                                udp_port=udp_port)
                session.add(vxlan_endpoint)
            return vxlan_endpoint



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013 Embrane, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Ivar Lazzaro, Embrane, Inc.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class SupportBase(object):
    """abstract support class.

    Defines the methods a plugin support should implement to be used as
    the L2 base for Embrane plugin.

    """

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def retrieve_utif_info(self, context, neutron_port=None, network=None):
        """Retrieve specific network info.

        each plugin support, querying its own DB, can collect all the
        information needed by the ESM in order to create the
        user traffic security zone.

        :param interface_info: the foo parameter
        :param context: neutron request context
        :returns: heleosapi.info.UtifInfo -- specific network info
        :raises: UtifInfoError
        """



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sqlalchemy as sa

from neutron.db import db_base_plugin_v2
from neutron.db import l3_db
from neutron.extensions import l3
from neutron.openstack.common import log as logging


LOG = logging.getLogger(__name__)
EXTERNAL_GW_INFO = l3.EXTERNAL_GW_INFO

# Modify the Router Data Model adding the enable_snat attribute
setattr(l3_db.Router, 'enable_snat',
        sa.Column(sa.Boolean, default=True, nullable=False))


class L3_NAT_db_mixin(l3_db.L3_NAT_db_mixin):
    """Mixin class to add configurable gateway modes."""

    # Register dict extend functions for ports and networks
    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
        l3.ROUTERS, ['_extend_router_dict_gw_mode'])

    def _extend_router_dict_gw_mode(self, router_res, router_db):
        if router_db.gw_port_id:
            nw_id = router_db.gw_port['network_id']
            router_res[EXTERNAL_GW_INFO] = {
                'network_id': nw_id,
                'enable_snat': router_db.enable_snat}

    def _update_router_gw_info(self, context, router_id, info, router=None):
        # Load the router only if necessary
        if not router:
            router = self._get_router(context, router_id)
        # if enable_snat is not specified use the value
        # stored in the database (default:True)
        enable_snat = not info or info.get('enable_snat', router.enable_snat)
        with context.session.begin(subtransactions=True):
            router.enable_snat = enable_snat

        # Calls superclass, pass router db object for avoiding re-loading
        super(L3_NAT_db_mixin, self)._update_router_gw_info(
            context, router_id, info, router=router)
        # Returning the router might come back useful if this
        # method is overriden in child classes
        return router

    def _build_routers_list(self, routers, gw_ports):
        gw_port_id_gw_port_dict = {}
        for gw_port in gw_ports:
            gw_port_id_gw_port_dict[gw_port['id']] = gw_port
        for rtr in routers:
            gw_port_id = rtr['gw_port_id']
            if gw_port_id:
                rtr['gw_port'] = gw_port_id_gw_port_dict[gw_port_id]
                # Add enable_snat key
                rtr['enable_snat'] = rtr[EXTERNAL_GW_INFO]['enable_snat']
        return routers



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Paul Michali, Cisco Systems, Inc.

import abc
from collections import namedtuple
import requests

import netaddr
from oslo.config import cfg

from neutron.common import exceptions
from neutron.common import rpc as n_rpc
from neutron import context as ctx
from neutron.openstack.common import lockutils
from neutron.openstack.common import log as logging
from neutron.openstack.common import loopingcall
from neutron.openstack.common import rpc
from neutron.openstack.common.rpc import proxy
from neutron.plugins.common import constants
from neutron.plugins.common import utils as plugin_utils
from neutron.services.vpn.common import topics
from neutron.services.vpn import device_drivers
from neutron.services.vpn.device_drivers import (
    cisco_csr_rest_client as csr_client)


ipsec_opts = [
    cfg.IntOpt('status_check_interval',
               default=60,
               help=_("Status check interval for Cisco CSR IPSec connections"))
]
cfg.CONF.register_opts(ipsec_opts, 'cisco_csr_ipsec')

LOG = logging.getLogger(__name__)

RollbackStep = namedtuple('RollbackStep', ['action', 'resource_id', 'title'])


class CsrResourceCreateFailure(exceptions.NeutronException):
    message = _("Cisco CSR failed to create %(resource)s (%(which)s)")


class CsrDriverMismatchError(exceptions.NeutronException):
    message = _("Required %(resource)s attribute %(attr)s mapping for Cisco "
                "CSR is missing in device driver")


class CsrUnknownMappingError(exceptions.NeutronException):
    message = _("Device driver does not have a mapping of '%(value)s for "
                "attribute %(attr)s of %(resource)s")


def find_available_csrs_from_config(config_files):
    """Read INI for available Cisco CSRs that driver can use.

    Loads management port, tunnel IP, user, and password information for
    available CSRs from configuration file. Driver will use this info to
    configure VPN connections. The CSR is associated 1:1 with a Neutron
    router. To identify which CSR to use for a VPN service, the public
    (GW) IP of the Neutron router will be used as an index into the CSR
    config info.
    """
    multi_parser = cfg.MultiConfigParser()
    LOG.info(_("Scanning config files %s for Cisco CSR configurations"),
             config_files)
    try:
        read_ok = multi_parser.read(config_files)
    except cfg.ParseError as pe:
        LOG.error(_("Config file parse error: %s"), pe)
        return {}

    if len(read_ok) != len(config_files):
        raise cfg.Error(_("Unable to parse config files %s for Cisco CSR "
                          "info") % config_files)
    csrs_found = {}
    for parsed_file in multi_parser.parsed:
        for parsed_item in parsed_file.keys():
            device_type, sep, for_router = parsed_item.partition(':')
            if device_type.lower() == 'cisco_csr_rest':
                try:
                    netaddr.IPNetwork(for_router)
                except netaddr.core.AddrFormatError:
                    LOG.error(_("Ignoring Cisco CSR configuration entry - "
                                "router IP %s is not valid"), for_router)
                    continue
                entry = parsed_file[parsed_item]
                # Check for missing fields
                try:
                    rest_mgmt_ip = entry['rest_mgmt'][0]
                    tunnel_ip = entry['tunnel_ip'][0]
                    username = entry['username'][0]
                    password = entry['password'][0]
                except KeyError as ke:
                    LOG.error(_("Ignoring Cisco CSR for router %(router)s "
                                "- missing %(field)s setting"),
                              {'router': for_router, 'field': str(ke)})
                    continue
                # Validate fields
                try:
                    timeout = float(entry['timeout'][0])
                except ValueError:
                    LOG.error(_("Ignoring Cisco CSR for router %s - "
                                "timeout is not a floating point number"),
                              for_router)
                    continue
                except KeyError:
                    timeout = csr_client.TIMEOUT
                try:
                    netaddr.IPAddress(rest_mgmt_ip)
                except netaddr.core.AddrFormatError:
                    LOG.error(_("Ignoring Cisco CSR for subnet %s - "
                                "REST management is not an IP address"),
                              for_router)
                    continue
                try:
                    netaddr.IPAddress(tunnel_ip)
                except netaddr.core.AddrFormatError:
                    LOG.error(_("Ignoring Cisco CSR for router %s - "
                                "local tunnel is not an IP address"),
                              for_router)
                    continue
                csrs_found[for_router] = {'rest_mgmt': rest_mgmt_ip,
                                          'tunnel_ip': tunnel_ip,
                                          'username': username,
                                          'password': password,
                                          'timeout': timeout}

                LOG.debug(_("Found CSR for router %(router)s: %(info)s"),
                          {'router': for_router,
                           'info': csrs_found[for_router]})
    return csrs_found


class CiscoCsrIPsecVpnDriverApi(proxy.RpcProxy):
    """RPC API for agent to plugin messaging."""

    def get_vpn_services_on_host(self, context, host):
        """Get list of vpnservices on this host.

        The vpnservices including related ipsec_site_connection,
        ikepolicy, ipsecpolicy, and Cisco info on this host.
        """
        return self.call(context,
                         self.make_msg('get_vpn_services_on_host',
                                       host=host),
                         topic=self.topic)

    def update_status(self, context, status):
        """Update status for all VPN services and connections."""
        return self.cast(context,
                         self.make_msg('update_status',
                                       status=status),
                         topic=self.topic)


class CiscoCsrIPsecDriver(device_drivers.DeviceDriver):
    """Cisco CSR VPN Device Driver for IPSec.

    This class is designed for use with L3-agent now.
    However this driver will be used with another agent in future.
    so the use of "Router" is kept minimul now.
    Insted of router_id,  we are using process_id in this code.
    """

    # history
    #   1.0 Initial version

    RPC_API_VERSION = '1.0'
    __metaclass__ = abc.ABCMeta

    def __init__(self, agent, host):
        self.host = host
        self.conn = rpc.create_connection(new=True)
        context = ctx.get_admin_context_without_session()
        node_topic = '%s.%s' % (topics.CISCO_IPSEC_AGENT_TOPIC, self.host)

        self.service_state = {}

        self.conn.create_consumer(
            node_topic,
            self.create_rpc_dispatcher(),
            fanout=False)
        self.conn.consume_in_thread()
        self.agent_rpc = (
            CiscoCsrIPsecVpnDriverApi(topics.CISCO_IPSEC_DRIVER_TOPIC, '1.0'))
        self.periodic_report = loopingcall.FixedIntervalLoopingCall(
            self.report_status, context)
        self.periodic_report.start(
            interval=agent.conf.cisco_csr_ipsec.status_check_interval)

        csrs_found = find_available_csrs_from_config(cfg.CONF.config_file)
        if csrs_found:
            LOG.info(_("Loaded %(num)d Cisco CSR configuration%(plural)s"),
                     {'num': len(csrs_found),
                      'plural': 's'[len(csrs_found) == 1:]})
        else:
            raise SystemExit(_('No Cisco CSR configurations found in: %s') %
                             cfg.CONF.config_file)
        self.csrs = dict([(k, csr_client.CsrRestClient(v['rest_mgmt'],
                                                       v['tunnel_ip'],
                                                       v['username'],
                                                       v['password'],
                                                       v['timeout']))
                          for k, v in csrs_found.items()])

    def create_rpc_dispatcher(self):
        return n_rpc.PluginRpcDispatcher([self])

    def vpnservice_updated(self, context, **kwargs):
        """Handle VPNaaS service driver change notifications."""
        LOG.debug(_("Handling VPN service update notification '%s'"),
                  kwargs.get('reason', ''))
        self.sync(context, [])

    def create_vpn_service(self, service_data):
        """Create new entry to track VPN service and its connections."""
        vpn_service_id = service_data['id']
        vpn_service_router = service_data['external_ip']
        self.service_state[vpn_service_id] = CiscoCsrVpnService(
            service_data, self.csrs.get(vpn_service_router))
        return self.service_state[vpn_service_id]

    def update_connection(self, context, vpn_service_id, conn_data):
        """Handle notification for a single IPSec connection."""
        vpn_service = self.service_state[vpn_service_id]
        conn_id = conn_data['id']
        conn_is_admin_up = conn_data[u'admin_state_up']

        if conn_id in vpn_service.conn_state:
            ipsec_conn = vpn_service.conn_state[conn_id]
            if ipsec_conn.forced_down:
                if vpn_service.is_admin_up and conn_is_admin_up:
                    LOG.debug(_("Update: Connection %s no longer admin down"),
                              conn_id)
                    # TODO(pcm) Do no shut on tunnel, once CSR supports
                    ipsec_conn.forced_down = False
                    ipsec_conn.create_ipsec_site_connection(context, conn_data)
            else:
                if not vpn_service.is_admin_up or not conn_is_admin_up:
                    LOG.debug(_("Update: Connection %s forced to admin down"),
                              conn_id)
                    # TODO(pcm) Do shut on tunnel, once CSR supports
                    ipsec_conn.forced_down = True
                    ipsec_conn.delete_ipsec_site_connection(context, conn_id)
                else:
                    # TODO(pcm) FUTURE handle connection update
                    LOG.debug(_("Update: Ignoring existing connection %s"),
                              conn_id)
        else:  # New connection...
            ipsec_conn = vpn_service.create_connection(conn_data)
            if not vpn_service.is_admin_up or not conn_is_admin_up:
                # TODO(pcm) Create, but set tunnel down, once CSR supports
                LOG.debug(_("Update: Created new connection %s in admin down "
                            "state"), conn_id)
                ipsec_conn.forced_down = True
            else:
                LOG.debug(_("Update: Created new connection %s"), conn_id)
                ipsec_conn.create_ipsec_site_connection(context, conn_data)

        ipsec_conn.is_dirty = False
        ipsec_conn.last_status = conn_data['status']
        ipsec_conn.is_admin_up = conn_is_admin_up
        return ipsec_conn

    def update_service(self, context, service_data):
        """Handle notification for a single VPN Service and its connections."""
        vpn_service_id = service_data['id']
        csr_id = service_data['external_ip']
        if csr_id not in self.csrs:
            LOG.error(_("Update: Skipping VPN service %(service)s as it's "
                        "router (%(csr_id)s is not associated with a Cisco "
                        "CSR"), {'service': vpn_service_id, 'csr_id': csr_id})
            return

        if vpn_service_id in self.service_state:
            LOG.debug(_("Update: Existing VPN service %s detected"),
                      vpn_service_id)
            vpn_service = self.service_state[vpn_service_id]
        else:
            LOG.debug(_("Update: New VPN service %s detected"), vpn_service_id)
            vpn_service = self.create_vpn_service(service_data)

        vpn_service.is_dirty = False
        vpn_service.connections_removed = False
        vpn_service.last_status = service_data['status']
        vpn_service.is_admin_up = service_data[u'admin_state_up']
        for conn_data in service_data['ipsec_conns']:
            self.update_connection(context, vpn_service_id, conn_data)
        LOG.debug(_("Update: Completed update processing"))
        return vpn_service

    def update_all_services_and_connections(self, context):
        """Update services and connections based on plugin info.

        Perform any create and update operations and then update status.
        Mark every visited connection as no longer "dirty" so they will
        not be deleted at end of sync processing.
        """
        services_data = self.agent_rpc.get_vpn_services_on_host(context,
                                                                self.host)
        LOG.debug("Sync updating for %d VPN services", len(services_data))
        vpn_services = []
        for service_data in services_data:
            vpn_service = self.update_service(context, service_data)
            if vpn_service:
                vpn_services.append(vpn_service)
        return vpn_services

    def mark_existing_connections_as_dirty(self):
        """Mark all existing connections as "dirty" for sync."""
        service_count = 0
        connection_count = 0
        for service_state in self.service_state.values():
            service_state.is_dirty = True
            service_count += 1
            for conn_id in service_state.conn_state:
                service_state.conn_state[conn_id].is_dirty = True
                connection_count += 1
        LOG.debug(_("Mark: %(service)d VPN services and %(conn)d IPSec "
                    "connections marked dirty"), {'service': service_count,
                                                  'conn': connection_count})

    def remove_unknown_connections(self, context):
        """Remove connections that are not known by service driver."""
        service_count = 0
        connection_count = 0
        for vpn_service_id, vpn_service in self.service_state.items():
            dirty = [c_id for c_id, c in vpn_service.conn_state.items()
                     if c.is_dirty]
            vpn_service.connections_removed = len(dirty) > 0
            for conn_id in dirty:
                conn_state = vpn_service.conn_state[conn_id]
                conn_state.delete_ipsec_site_connection(context, conn_id)
                connection_count += 1
                del vpn_service.conn_state[conn_id]
            if vpn_service.is_dirty:
                service_count += 1
                del self.service_state[vpn_service_id]
            elif dirty:
                self.connections_removed = True
        LOG.debug(_("Sweep: Removed %(service)d dirty VPN service%(splural)s "
                    "and %(conn)d dirty IPSec connection%(cplural)s"),
                  {'service': service_count, 'conn': connection_count,
                   'splural': 's'[service_count == 1:],
                   'cplural': 's'[connection_count == 1:]})

    def build_report_for_connections_on(self, vpn_service):
        """Create the report fragment for IPSec connections on a service.

        Collect the current status from the Cisco CSR and use that to update
        the status and generate report fragment for each connection on the
        service. If there is no status information, or no change, then no
        report info will be created for the connection. The combined report
        data is returned.
        """
        LOG.debug(_("Report: Collecting status for IPSec connections on VPN "
                    "service %s"), vpn_service.service_id)
        tunnels = vpn_service.get_ipsec_connections_status()
        report = {}
        for connection in vpn_service.conn_state.values():
            if connection.forced_down:
                LOG.debug(_("Connection %s forced down"), connection.conn_id)
                current_status = constants.DOWN
            else:
                current_status = connection.find_current_status_in(tunnels)
                LOG.debug(_("Connection %(conn)s reported %(status)s"),
                          {'conn': connection.conn_id,
                           'status': current_status})
            frag = connection.update_status_and_build_report(current_status)
            if frag:
                LOG.debug(_("Report: Adding info for IPSec connection %s"),
                          connection.conn_id)
                report.update(frag)
        return report

    def build_report_for_service(self, vpn_service):
        """Create the report info for a VPN service and its IPSec connections.

        Get the report info for the connections on the service, and include
        it into the report info for the VPN service. If there is no report
        info for the connection, then no change has occurred and no report
        will be generated. If there is only one connection for the service,
        we'll set the service state to match the connection (with ERROR seen
        as DOWN).
        """
        conn_report = self.build_report_for_connections_on(vpn_service)
        if conn_report or vpn_service.connections_removed:
            pending_handled = plugin_utils.in_pending_status(
                vpn_service.last_status)
            vpn_service.update_last_status()
            LOG.debug(_("Report: Adding info for VPN service %s"),
                      vpn_service.service_id)
            return {u'id': vpn_service.service_id,
                    u'status': vpn_service.last_status,
                    u'updated_pending_status': pending_handled,
                    u'ipsec_site_connections': conn_report}
        else:
            return {}

    @lockutils.synchronized('vpn-agent', 'neutron-')
    def report_status(self, context):
        """Report status of all VPN services and IPSec connections to plugin.

        This is called periodically by the agent, to push up changes in
        status. Use a lock to serialize access to (and changing of)
        running state.
        """
        return self.report_status_internal(context)

    def report_status_internal(self, context):
        """Generate report and send to plugin, if anything changed."""
        service_report = []
        LOG.debug(_("Report: Starting status report processing"))
        for vpn_service_id, vpn_service in self.service_state.items():
            LOG.debug(_("Report: Collecting status for VPN service %s"),
                      vpn_service_id)
            report = self.build_report_for_service(vpn_service)
            if report:
                service_report.append(report)
        if service_report:
            LOG.info(_("Sending status report update to plugin"))
            self.agent_rpc.update_status(context, service_report)
        LOG.debug(_("Report: Completed status report processing"))
        return service_report

    @lockutils.synchronized('vpn-agent', 'neutron-')
    def sync(self, context, routers):
        """Synchronize with plugin and report current status.

        Mark all "known" services/connections as dirty, update them based on
        information from the plugin, remove (sweep) any connections that are
        not updated (dirty), and report updates, if any, back to plugin.
        Called when update/delete a service or create/update/delete a
        connection (vpnservice_updated message), or router change
        (_process_routers).

        Use lock to serialize access (and changes) to running state for VPN
        service and IPsec connections.
        """
        self.mark_existing_connections_as_dirty()
        self.update_all_services_and_connections(context)
        self.remove_unknown_connections(context)
        self.report_status_internal(context)

    def create_router(self, process_id):
        """Actions taken when router created."""
        # Note: Since Cisco CSR is running out-of-band, nothing to do here
        pass

    def destroy_router(self, process_id):
        """Actions taken when router deleted."""
        # Note: Since Cisco CSR is running out-of-band, nothing to do here
        pass


class CiscoCsrVpnService(object):

    """Maintains state/status information for a service and its connections."""

    def __init__(self, service_data, csr):
        self.service_id = service_data['id']
        self.conn_state = {}
        self.csr = csr
        self.is_admin_up = True
        # TODO(pcm) FUTURE - handle sharing of policies

    def create_connection(self, conn_data):
        conn_id = conn_data['id']
        self.conn_state[conn_id] = CiscoCsrIPSecConnection(conn_data, self.csr)
        return self.conn_state[conn_id]

    def get_connection(self, conn_id):
        return self.conn_state.get(conn_id)

    def conn_status(self, conn_id):
        conn_state = self.get_connection(conn_id)
        if conn_state:
            return conn_state.last_status

    def snapshot_conn_state(self, ipsec_conn):
        """Create/obtain connection state and save current status."""
        conn_state = self.conn_state.setdefault(
            ipsec_conn['id'], CiscoCsrIPSecConnection(ipsec_conn, self.csr))
        conn_state.last_status = ipsec_conn['status']
        conn_state.is_dirty = False
        return conn_state

    STATUS_MAP = {'ERROR': constants.ERROR,
                  'UP-ACTIVE': constants.ACTIVE,
                  'UP-IDLE': constants.ACTIVE,
                  'UP-NO-IKE': constants.ACTIVE,
                  'DOWN': constants.DOWN,
                  'DOWN-NEGOTIATING': constants.DOWN}

    def get_ipsec_connections_status(self):
        """Obtain current status of all tunnels on a Cisco CSR.

        Convert them to OpenStack status values.
        """
        tunnels = self.csr.read_tunnel_statuses()
        for tunnel in tunnels:
            LOG.debug("CSR Reports %(tunnel)s status '%(status)s'",
                      {'tunnel': tunnel[0], 'status': tunnel[1]})
        return dict(map(lambda x: (x[0], self.STATUS_MAP[x[1]]), tunnels))

    def find_matching_connection(self, tunnel_id):
        """Find IPSec connection using Cisco CSR tunnel specified, if any."""
        for connection in self.conn_state.values():
            if connection.tunnel == tunnel_id:
                return connection.conn_id

    def no_connections_up(self):
        return not any(c.last_status == 'ACTIVE'
                       for c in self.conn_state.values())

    def update_last_status(self):
        if not self.is_admin_up or self.no_connections_up():
            self.last_status = constants.DOWN
        else:
            self.last_status = constants.ACTIVE


class CiscoCsrIPSecConnection(object):

    """State and actions for IPSec site-to-site connections."""

    def __init__(self, conn_info, csr):
        self.conn_id = conn_info['id']
        self.csr = csr
        self.steps = []
        self.forced_down = False
        self.is_admin_up = conn_info[u'admin_state_up']
        self.tunnel = conn_info['cisco']['site_conn_id']

    def find_current_status_in(self, statuses):
        if self.tunnel in statuses:
            return statuses[self.tunnel]
        else:
            return constants.ERROR

    def update_status_and_build_report(self, current_status):
        if current_status != self.last_status:
            pending_handled = plugin_utils.in_pending_status(self.last_status)
            self.last_status = current_status
            return {self.conn_id: {'status': current_status,
                                   'updated_pending_status': pending_handled}}
        else:
            return {}

    DIALECT_MAP = {'ike_policy': {'name': 'IKE Policy',
                                  'v1': u'v1',
                                  # auth_algorithm -> hash
                                  'sha1': u'sha',
                                  # encryption_algorithm -> encryption
                                  '3des': u'3des',
                                  'aes-128': u'aes',
                                  # TODO(pcm) update these 2 once CSR updated
                                  'aes-192': u'aes',
                                  'aes-256': u'aes',
                                  # pfs -> dhGroup
                                  'group2': 2,
                                  'group5': 5,
                                  'group14': 14},
                   'ipsec_policy': {'name': 'IPSec Policy',
                                    # auth_algorithm -> esp-authentication
                                    'sha1': u'esp-sha-hmac',
                                    # transform_protocol -> ah
                                    'esp': None,
                                    'ah': u'ah-sha-hmac',
                                    'ah-esp': u'ah-sha-hmac',
                                    # encryption_algorithm -> esp-encryption
                                    '3des': u'esp-3des',
                                    'aes-128': u'esp-aes',
                                    # TODO(pcm) update these 2 once CSR updated
                                    'aes-192': u'esp-aes',
                                    'aes-256': u'esp-aes',
                                    # pfs -> pfs
                                    'group2': u'group2',
                                    'group5': u'group5',
                                    'group14': u'group14'}}

    def translate_dialect(self, resource, attribute, info):
        """Map VPNaaS attributes values to CSR values for a resource."""
        name = self.DIALECT_MAP[resource]['name']
        if attribute not in info:
            raise CsrDriverMismatchError(resource=name, attr=attribute)
        value = info[attribute].lower()
        if value in self.DIALECT_MAP[resource]:
            return self.DIALECT_MAP[resource][value]
        raise CsrUnknownMappingError(resource=name, attr=attribute,
                                     value=value)

    def create_psk_info(self, psk_id, conn_info):
        """Collect/create attributes needed for pre-shared key."""
        return {u'keyring-name': psk_id,
                u'pre-shared-key-list': [
                    {u'key': conn_info['psk'],
                     u'encrypted': False,
                     u'peer-address': conn_info['peer_address']}]}

    def create_ike_policy_info(self, ike_policy_id, conn_info):
        """Collect/create/map attributes needed for IKE policy."""
        for_ike = 'ike_policy'
        policy_info = conn_info[for_ike]
        version = self.translate_dialect(for_ike,
                                         'ike_version',
                                         policy_info)
        encrypt_algorithm = self.translate_dialect(for_ike,
                                                   'encryption_algorithm',
                                                   policy_info)
        auth_algorithm = self.translate_dialect(for_ike,
                                                'auth_algorithm',
                                                policy_info)
        group = self.translate_dialect(for_ike,
                                       'pfs',
                                       policy_info)
        lifetime = policy_info['lifetime_value']
        return {u'version': version,
                u'priority-id': ike_policy_id,
                u'encryption': encrypt_algorithm,
                u'hash': auth_algorithm,
                u'dhGroup': group,
                u'lifetime': lifetime}

    def create_ipsec_policy_info(self, ipsec_policy_id, info):
        """Collect/create attributes needed for IPSec policy.

        Note: OpenStack will provide a default encryption algorithm, if one is
        not provided, so a authentication only configuration of (ah, sha1),
        which maps to ah-sha-hmac transform protocol, cannot be selected.
        As a result, we'll always configure the encryption algorithm, and
        will select ah-sha-hmac for transform protocol.
        """

        for_ipsec = 'ipsec_policy'
        policy_info = info[for_ipsec]
        transform_protocol = self.translate_dialect(for_ipsec,
                                                    'transform_protocol',
                                                    policy_info)
        auth_algorithm = self.translate_dialect(for_ipsec,
                                                'auth_algorithm',
                                                policy_info)
        encrypt_algorithm = self.translate_dialect(for_ipsec,
                                                   'encryption_algorithm',
                                                   policy_info)
        group = self.translate_dialect(for_ipsec, 'pfs', policy_info)
        lifetime = policy_info['lifetime_value']
        settings = {u'policy-id': ipsec_policy_id,
                    u'protection-suite': {
                        u'esp-encryption': encrypt_algorithm,
                        u'esp-authentication': auth_algorithm},
                    u'lifetime-sec': lifetime,
                    u'pfs': group,
                    # TODO(pcm): Remove when CSR fixes 'Disable'
                    u'anti-replay-window-size': u'64'}
        if transform_protocol:
            settings[u'protection-suite'][u'ah'] = transform_protocol
        return settings

    def create_site_connection_info(self, site_conn_id, ipsec_policy_id,
                                    conn_info):
        """Collect/create attributes needed for the IPSec connection."""
        # TODO(pcm) Enable, once CSR is embedded as a Neutron router
        # gw_ip = vpnservice['external_ip'] (need to pass in)
        mtu = conn_info['mtu']
        return {
            u'vpn-interface-name': site_conn_id,
            u'ipsec-policy-id': ipsec_policy_id,
            u'local-device': {
                # TODO(pcm): FUTURE - Get CSR port of interface with
                # local subnet
                u'ip-address': u'GigabitEthernet3',
                # TODO(pcm): FUTURE - Get IP address of router's public
                # I/F, once CSR is used as embedded router.
                u'tunnel-ip-address': u'172.24.4.23'
                # u'tunnel-ip-address': u'%s' % gw_ip
            },
            u'remote-device': {
                u'tunnel-ip-address': conn_info['peer_address']
            },
            u'mtu': mtu
        }

    def create_routes_info(self, site_conn_id, conn_info):
        """Collect/create attributes for static routes."""
        routes_info = []
        for peer_cidr in conn_info.get('peer_cidrs', []):
            route = {u'destination-network': peer_cidr,
                     u'outgoing-interface': site_conn_id}
            route_id = csr_client.make_route_id(peer_cidr, site_conn_id)
            routes_info.append((route_id, route))
        return routes_info

    def _check_create(self, resource, which):
        """Determine if REST create request was successful."""
        if self.csr.status == requests.codes.CREATED:
            LOG.debug("%(resource)s %(which)s is configured",
                      {'resource': resource, 'which': which})
            return
        LOG.error(_("Unable to create %(resource)s %(which)s: "
                    "%(status)d"),
                  {'resource': resource, 'which': which,
                   'status': self.csr.status})
        # ToDO(pcm): Set state to error
        raise CsrResourceCreateFailure(resource=resource, which=which)

    def do_create_action(self, action_suffix, info, resource_id, title):
        """Perform a single REST step for IPSec site connection create."""
        create_action = 'create_%s' % action_suffix
        try:
            getattr(self.csr, create_action)(info)
        except AttributeError:
            LOG.exception(_("Internal error - '%s' is not defined"),
                          create_action)
            raise CsrResourceCreateFailure(resource=title,
                                           which=resource_id)
        self._check_create(title, resource_id)
        self.steps.append(RollbackStep(action_suffix, resource_id, title))

    def _verify_deleted(self, status, resource, which):
        """Determine if REST delete request was successful."""
        if status in (requests.codes.NO_CONTENT, requests.codes.NOT_FOUND):
            LOG.debug("%(resource)s configuration %(which)s was removed",
                      {'resource': resource, 'which': which})
        else:
            LOG.warning(_("Unable to delete %(resource)s %(which)s: "
                          "%(status)d"), {'resource': resource,
                                          'which': which,
                                          'status': status})

    def do_rollback(self):
        """Undo create steps that were completed successfully."""
        for step in reversed(self.steps):
            delete_action = 'delete_%s' % step.action
            LOG.debug(_("Performing rollback action %(action)s for "
                        "resource %(resource)s"), {'action': delete_action,
                                                   'resource': step.title})
            try:
                getattr(self.csr, delete_action)(step.resource_id)
            except AttributeError:
                LOG.exception(_("Internal error - '%s' is not defined"),
                              delete_action)
                raise CsrResourceCreateFailure(resource=step.title,
                                               which=step.resource_id)
            self._verify_deleted(self.csr.status, step.title, step.resource_id)
        self.steps = []

    def create_ipsec_site_connection(self, context, conn_info):
        """Creates an IPSec site-to-site connection on CSR.

        Create the PSK, IKE policy, IPSec policy, connection, static route,
        and (future) DPD.
        """
        # Get all the IDs
        conn_id = conn_info['id']
        psk_id = conn_id
        site_conn_id = conn_info['cisco']['site_conn_id']
        ike_policy_id = conn_info['cisco']['ike_policy_id']
        ipsec_policy_id = conn_info['cisco']['ipsec_policy_id']

        LOG.debug(_('Creating IPSec connection %s'), conn_id)
        # Get all the attributes needed to create
        try:
            psk_info = self.create_psk_info(psk_id, conn_info)
            ike_policy_info = self.create_ike_policy_info(ike_policy_id,
                                                          conn_info)
            ipsec_policy_info = self.create_ipsec_policy_info(ipsec_policy_id,
                                                              conn_info)
            connection_info = self.create_site_connection_info(site_conn_id,
                                                               ipsec_policy_id,
                                                               conn_info)
            routes_info = self.create_routes_info(site_conn_id, conn_info)
        except (CsrUnknownMappingError, CsrDriverMismatchError) as e:
            LOG.exception(e)
            return

        try:
            self.do_create_action('pre_shared_key', psk_info,
                                  conn_id, 'Pre-Shared Key')
            self.do_create_action('ike_policy', ike_policy_info,
                                  ike_policy_id, 'IKE Policy')
            self.do_create_action('ipsec_policy', ipsec_policy_info,
                                  ipsec_policy_id, 'IPSec Policy')
            self.do_create_action('ipsec_connection', connection_info,
                                  site_conn_id, 'IPSec Connection')

            # TODO(pcm): FUTURE - Do DPD for v1 and handle if >1 connection
            # and different DPD settings
            for route_id, route_info in routes_info:
                self.do_create_action('static_route', route_info,
                                      route_id, 'Static Route')
        except CsrResourceCreateFailure:
            self.do_rollback()
            LOG.info(_("FAILED: Create of IPSec site-to-site connection %s"),
                     conn_id)
        else:
            LOG.info(_("SUCCESS: Created IPSec site-to-site connection %s"),
                     conn_id)

    def delete_ipsec_site_connection(self, context, conn_id):
        """Delete the site-to-site IPSec connection.

        This will be best effort and will continue, if there are any
        failures.
        """
        LOG.debug(_('Deleting IPSec connection %s'), conn_id)
        if not self.steps:
            LOG.warning(_('Unable to find connection %s'), conn_id)
        else:
            self.do_rollback()

        LOG.info(_("SUCCESS: Deleted IPSec site-to-site connection %s"),
                 conn_id)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock


from neutron import context as n_ctx
from neutron.db import api as dbapi
from neutron.openstack.common import uuidutils
from neutron.plugins.common import constants
from neutron.services.vpn.service_drivers import cisco_csr_db as csr_db
from neutron.services.vpn.service_drivers import cisco_ipsec as ipsec_driver
from neutron.tests import base

_uuid = uuidutils.generate_uuid

FAKE_VPN_CONN_ID = _uuid()

FAKE_VPN_CONNECTION = {
    'vpnservice_id': _uuid(),
    'id': FAKE_VPN_CONN_ID,
    'ikepolicy_id': _uuid(),
    'ipsecpolicy_id': _uuid(),
    'tenant_id': _uuid()
}
FAKE_VPN_SERVICE = {
    'router_id': _uuid()
}
FAKE_HOST = 'fake_host'


class TestCiscoIPsecDriverValidation(base.BaseTestCase):

    def setUp(self):
        super(TestCiscoIPsecDriverValidation, self).setUp()
        mock.patch('neutron.openstack.common.rpc.create_connection').start()
        self.service_plugin = mock.Mock()
        self.driver = ipsec_driver.CiscoCsrIPsecVPNDriver(self.service_plugin)
        self.context = n_ctx.Context('some_user', 'some_tenant')
        self.vpn_service = mock.Mock()

    def test_ike_version_unsupported(self):
        """Failure test that Cisco CSR REST API does not support IKE v2."""
        policy_info = {'ike_version': 'v2',
                       'lifetime': {'units': 'seconds', 'value': 60}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_ike_version, policy_info)

    def test_ike_lifetime_not_in_seconds(self):
        """Failure test of unsupported lifetime units for IKE policy."""
        policy_info = {'lifetime': {'units': 'kilobytes', 'value': 1000}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          "IKE Policy", policy_info)

    def test_ipsec_lifetime_not_in_seconds(self):
        """Failure test of unsupported lifetime units for IPSec policy."""
        policy_info = {'lifetime': {'units': 'kilobytes', 'value': 1000}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          "IPSec Policy", policy_info)

    def test_ike_lifetime_seconds_values_at_limits(self):
        """Test valid lifetime values for IKE policy."""
        policy_info = {'lifetime': {'units': 'seconds', 'value': 60}}
        self.driver.validate_lifetime('IKE Policy', policy_info)
        policy_info = {'lifetime': {'units': 'seconds', 'value': 86400}}
        self.driver.validate_lifetime('IKE Policy', policy_info)

    def test_ipsec_lifetime_seconds_values_at_limits(self):
        """Test valid lifetime values for IPSec policy."""
        policy_info = {'lifetime': {'units': 'seconds', 'value': 120}}
        self.driver.validate_lifetime('IPSec Policy', policy_info)
        policy_info = {'lifetime': {'units': 'seconds', 'value': 2592000}}
        self.driver.validate_lifetime('IPSec Policy', policy_info)

    def test_ike_lifetime_values_invalid(self):
        """Failure test of unsupported lifetime values for IKE policy."""
        which = "IKE Policy"
        policy_info = {'lifetime': {'units': 'seconds', 'value': 59}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          which, policy_info)
        policy_info = {'lifetime': {'units': 'seconds', 'value': 86401}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          which, policy_info)

    def test_ipsec_lifetime_values_invalid(self):
        """Failure test of unsupported lifetime values for IPSec policy."""
        which = "IPSec Policy"
        policy_info = {'lifetime': {'units': 'seconds', 'value': 119}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          which, policy_info)
        policy_info = {'lifetime': {'units': 'seconds', 'value': 2592001}}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_lifetime,
                          which, policy_info)

    def test_ipsec_connection_with_mtu_at_limits(self):
        """Test IPSec site-to-site connection with MTU at limits."""
        conn_info = {'mtu': 1500}
        self.driver.validate_mtu(conn_info)
        conn_info = {'mtu': 9192}
        self.driver.validate_mtu(conn_info)

    def test_ipsec_connection_with_invalid_mtu(self):
        """Failure test of IPSec site connection with unsupported MTUs."""
        conn_info = {'mtu': 1499}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_mtu, conn_info)
        conn_info = {'mtu': 9193}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_mtu, conn_info)

    def simulate_gw_ip_available(self):
        """Helper function indicating that tunnel has a gateway IP."""
        def have_one():
            return 1
        self.vpn_service.router.gw_port.fixed_ips.__len__ = have_one
        ip_addr_mock = mock.Mock()
        self.vpn_service.router.gw_port.fixed_ips = [ip_addr_mock]
        return ip_addr_mock

    def test_have_public_ip_for_router(self):
        """Ensure that router for IPSec connection has gateway IP."""
        self.simulate_gw_ip_available()
        self.driver.validate_public_ip_present(self.vpn_service)

    def test_router_with_missing_gateway_ip(self):
        """Failure test of IPSec connection with missing gateway IP."""
        self.simulate_gw_ip_available()
        self.vpn_service.router.gw_port = None
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_public_ip_present,
                          self.vpn_service)

    def test_peer_id_is_an_ip_address(self):
        """Ensure peer ID is an IP address for IPsec connection create."""
        ipsec_conn = {'peer_id': '10.10.10.10'}
        self.driver.validate_peer_id(ipsec_conn)

    def test_peer_id_is_not_ip_address(self):
        """Failure test of peer_id that is not an IP address."""
        ipsec_conn = {'peer_id': 'some-site.com'}
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.validate_peer_id, ipsec_conn)

    def test_validation_for_create_ipsec_connection(self):
        """Ensure all validation passes for IPSec site connection create."""
        self.simulate_gw_ip_available()
        # Provide the minimum needed items to validate
        ipsec_conn = {'id': '1',
                      'ikepolicy_id': '123',
                      'ipsecpolicy_id': '2',
                      'mtu': 1500,
                      'peer_id': '10.10.10.10'}
        self.service_plugin.get_ikepolicy = mock.Mock(
            return_value={'ike_version': 'v1',
                          'lifetime': {'units': 'seconds', 'value': 60}})
        self.service_plugin.get_ipsecpolicy = mock.Mock(
            return_value={'lifetime': {'units': 'seconds', 'value': 120}})
        self.driver.validate_ipsec_connection(self.context, ipsec_conn,
                                              self.vpn_service)


class TestCiscoIPsecDriverMapping(base.BaseTestCase):

    def setUp(self):
        super(TestCiscoIPsecDriverMapping, self).setUp()
        self.context = mock.patch.object(n_ctx, 'Context').start()
        self.session = self.context.session
        self.query_mock = self.session.query.return_value.order_by

    def test_identifying_first_mapping_id(self):
        """Make sure first available ID is obtained for each ID type."""
        # Simulate mapping table is empty - get first one
        self.query_mock.return_value = []
        next_id = csr_db.get_next_available_tunnel_id(self.session)
        self.assertEqual(0, next_id)

        next_id = csr_db.get_next_available_ike_policy_id(self.session)
        self.assertEqual(1, next_id)

        next_id = csr_db.get_next_available_ipsec_policy_id(self.session)
        self.assertEqual(1, next_id)

    def test_last_mapping_id_available(self):
        """Make sure can get the last ID for each of the table types."""
        # Simulate query indicates table is full
        self.query_mock.return_value = [
            (x, ) for x in xrange(csr_db.MAX_CSR_TUNNELS - 1)]
        next_id = csr_db.get_next_available_tunnel_id(self.session)
        self.assertEqual(csr_db.MAX_CSR_TUNNELS - 1, next_id)

        self.query_mock.return_value = [
            (x, ) for x in xrange(1, csr_db.MAX_CSR_IKE_POLICIES)]
        next_id = csr_db.get_next_available_ike_policy_id(self.session)
        self.assertEqual(csr_db.MAX_CSR_IKE_POLICIES, next_id)

        self.query_mock.return_value = [
            (x, ) for x in xrange(1, csr_db.MAX_CSR_IPSEC_POLICIES)]
        next_id = csr_db.get_next_available_ipsec_policy_id(self.session)
        self.assertEqual(csr_db.MAX_CSR_IPSEC_POLICIES, next_id)

    def test_reusing_first_available_mapping_id(self):
        """Ensure that we reuse the first available ID.

        Make sure that the next lowest ID is obtained from the mapping
        table when there are "holes" from deletions. Database query sorts
        the entries, so will return them in order. Using tunnel ID, as the
        logic is the same for each ID type.
        """
        self.query_mock.return_value = [(0, ), (1, ), (2, ), (5, ), (6, )]
        next_id = csr_db.get_next_available_tunnel_id(self.session)
        self.assertEqual(3, next_id)

    def test_no_more_mapping_ids_available(self):
        """Failure test of trying to reserve ID, when none available."""
        self.query_mock.return_value = [
            (x, ) for x in xrange(csr_db.MAX_CSR_TUNNELS)]
        self.assertRaises(IndexError, csr_db.get_next_available_tunnel_id,
                          self.session)

        self.query_mock.return_value = [
            (x, ) for x in xrange(1, csr_db.MAX_CSR_IKE_POLICIES + 1)]
        self.assertRaises(IndexError, csr_db.get_next_available_ike_policy_id,
                          self.session)

        self.query_mock.return_value = [
            (x, ) for x in xrange(1, csr_db.MAX_CSR_IPSEC_POLICIES + 1)]
        self.assertRaises(IndexError,
                          csr_db.get_next_available_ipsec_policy_id,
                          self.session)

    def test_create_tunnel_mappings(self):
        """Ensure successfully create new tunnel mappings."""
        # Simulate that first IDs are obtained
        self.query_mock.return_value = []
        map_db_mock = mock.patch.object(csr_db, 'IdentifierMap').start()
        conn_info = {'ikepolicy_id': '10',
                     'ipsecpolicy_id': '50',
                     'id': '100',
                     'tenant_id': '1000'}
        csr_db.create_tunnel_mapping(self.context, conn_info)
        map_db_mock.assert_called_once_with(csr_tunnel_id=0,
                                            csr_ike_policy_id=1,
                                            csr_ipsec_policy_id=1,
                                            ipsec_site_conn_id='100',
                                            tenant_id='1000')
        # Create another, with next ID of 2 for all IDs (not mocking each
        # ID separately, so will not have different IDs).
        self.query_mock.return_value = [(0, ), (1, )]
        map_db_mock.reset_mock()
        conn_info = {'ikepolicy_id': '20',
                     'ipsecpolicy_id': '60',
                     'id': '101',
                     'tenant_id': '1000'}
        csr_db.create_tunnel_mapping(self.context, conn_info)
        map_db_mock.assert_called_once_with(csr_tunnel_id=2,
                                            csr_ike_policy_id=2,
                                            csr_ipsec_policy_id=2,
                                            ipsec_site_conn_id='101',
                                            tenant_id='1000')


class TestCiscoIPsecDriver(base.BaseTestCase):

    """Test that various incoming requests are sent to device driver."""

    def setUp(self):
        super(TestCiscoIPsecDriver, self).setUp()
        dbapi.configure_db()
        self.addCleanup(dbapi.clear_db)
        mock.patch('neutron.openstack.common.rpc.create_connection').start()

        l3_agent = mock.Mock()
        l3_agent.host = FAKE_HOST
        plugin = mock.Mock()
        plugin.get_l3_agents_hosting_routers.return_value = [l3_agent]
        plugin_p = mock.patch('neutron.manager.NeutronManager.get_plugin')
        get_plugin = plugin_p.start()
        get_plugin.return_value = plugin
        service_plugin_p = mock.patch(
            'neutron.manager.NeutronManager.get_service_plugins')
        get_service_plugin = service_plugin_p.start()
        get_service_plugin.return_value = {constants.L3_ROUTER_NAT: plugin}

        service_plugin = mock.Mock()
        service_plugin.get_l3_agents_hosting_routers.return_value = [l3_agent]
        service_plugin._get_vpnservice.return_value = {
            'router_id': _uuid()
        }
        self.db_update_mock = service_plugin.update_ipsec_site_conn_status
        self.driver = ipsec_driver.CiscoCsrIPsecVPNDriver(service_plugin)
        self.driver.validate_ipsec_connection = mock.Mock()
        mock.patch.object(csr_db, 'create_tunnel_mapping').start()
        self.context = n_ctx.Context('some_user', 'some_tenant')

    def _test_update(self, func, args, reason=None):
        with mock.patch.object(self.driver.agent_rpc, 'cast') as cast:
            func(self.context, *args)
            cast.assert_called_once_with(
                self.context,
                {'args': reason,
                 'namespace': None,
                 'method': 'vpnservice_updated'},
                version='1.0',
                topic='cisco_csr_ipsec_agent.fake_host')

    def test_create_ipsec_site_connection(self):
        self._test_update(self.driver.create_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION],
                          {'reason': 'ipsec-conn-create'})

    def test_failure_validation_ipsec_connection(self):
        """Failure test of validation during IPSec site connection create.

        Simulate a validation failure, and ensure that database is
        updated to indicate connection is in error state.

        TODO(pcm): FUTURE - remove test case, once vendor plugin
        validation is done before database commit.
        """
        self.driver.validate_ipsec_connection.side_effect = (
            ipsec_driver.CsrValidationFailure(resource='IPSec Connection',
                                              key='mtu', value=1000))
        self.assertRaises(ipsec_driver.CsrValidationFailure,
                          self.driver.create_ipsec_site_connection,
                          self.context, FAKE_VPN_CONNECTION)
        self.db_update_mock.assert_called_with(self.context,
                                               FAKE_VPN_CONN_ID,
                                               constants.ERROR)

    def test_update_ipsec_site_connection(self):
        # TODO(pcm) FUTURE - Update test, when supported
        self.assertRaises(ipsec_driver.CsrUnsupportedError,
                          self._test_update,
                          self.driver.update_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION, FAKE_VPN_CONNECTION])

    def test_delete_ipsec_site_connection(self):
        self._test_update(self.driver.delete_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION],
                          {'reason': 'ipsec-conn-delete'})

    def test_update_vpnservice(self):
        self._test_update(self.driver.update_vpnservice,
                          [FAKE_VPN_SERVICE, FAKE_VPN_SERVICE],
                          {'reason': 'vpn-service-update'})

    def test_delete_vpnservice(self):
        self._test_update(self.driver.delete_vpnservice,
                          [FAKE_VPN_SERVICE],
                          {'reason': 'vpn-service-delete'})



# Copyright 2012, VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock
import testtools

from neutron.agent.linux import utils
from neutron.tests import base


class AgentUtilsExecuteTest(base.BaseTestCase):
    def setUp(self):
        super(AgentUtilsExecuteTest, self).setUp()
        self.root_helper = "echo"
        self.test_file = self.useFixture(
            fixtures.TempDir()).join("test_execute.tmp")
        open(self.test_file, 'w').close()
        self.mock_popen_p = mock.patch("subprocess.Popen.communicate")
        self.mock_popen = self.mock_popen_p.start()
        self.addCleanup(self.mock_popen_p.stop)

    def test_without_helper(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file])
        self.assertEqual(result, expected)

    def test_with_helper(self):
        expected = "ls %s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file],
                               self.root_helper)
        self.assertEqual(result, expected)

    def test_stderr_true(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        out = utils.execute(["ls", self.test_file], return_stderr=True)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (expected, ""))

    def test_check_exit_code(self):
        self.mock_popen.return_value = ["", ""]
        stdout = utils.execute(["ls", self.test_file[:-1]],
                               check_exit_code=False)
        self.assertEqual(stdout, "")

    def test_execute_raises(self):
        self.mock_popen.side_effect = RuntimeError
        self.assertRaises(RuntimeError, utils.execute,
                          ["ls", self.test_file[:-1]])

    def test_process_input(self):
        expected = "%s\n" % self.test_file[:-1]
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["cat"], process_input="%s\n" %
                               self.test_file[:-1])
        self.assertEqual(result, expected)

    def test_with_addl_env(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file],
                               addl_env={'foo': 'bar'})
        self.assertEqual(result, expected)


class AgentUtilsGetInterfaceMAC(base.BaseTestCase):
    def test_get_interface_mac(self):
        expect_val = '01:02:03:04:05:06'
        with mock.patch('fcntl.ioctl') as ioctl:
            ioctl.return_value = ''.join(['\x00' * 18,
                                          '\x01\x02\x03\x04\x05\x06',
                                          '\x00' * 232])
            actual_val = utils.get_interface_mac('eth0')
        self.assertEqual(actual_val, expect_val)


class AgentUtilsReplaceFile(base.BaseTestCase):
    def test_replace_file(self):
        # make file to replace
        with mock.patch('tempfile.NamedTemporaryFile') as ntf:
            ntf.return_value.name = '/baz'
            with mock.patch('os.chmod') as chmod:
                with mock.patch('os.rename') as rename:
                    utils.replace_file('/foo', 'bar')

                    expected = [mock.call('w+', dir='/', delete=False),
                                mock.call().write('bar'),
                                mock.call().close()]

                    ntf.assert_has_calls(expected)
                    chmod.assert_called_once_with('/baz', 0o644)
                    rename.assert_called_once_with('/baz', '/foo')


class TestFindChildPids(base.BaseTestCase):

    def test_returns_empty_list_for_exit_code_1(self):
        with mock.patch.object(utils, 'execute',
                               side_effect=RuntimeError('Exit code: 1')):
            self.assertEqual(utils.find_child_pids(-1), [])

    def test_returns_empty_list_for_no_output(self):
        with mock.patch.object(utils, 'execute', return_value=''):
            self.assertEqual(utils.find_child_pids(-1), [])

    def test_returns_list_of_child_process_ids_for_good_ouput(self):
        with mock.patch.object(utils, 'execute', return_value=' 123 \n 185\n'):
            self.assertEqual(utils.find_child_pids(-1), ['123', '185'])

    def test_raises_unknown_exception(self):
        with testtools.ExpectedException(RuntimeError):
            with mock.patch.object(utils, 'execute',
                                   side_effect=RuntimeError()):
                utils.find_child_pids(-1)



#!/usr/bin/env python
#
# Copyright (c) 2013 Brocade Communications Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Authors:
# Varma Bhupatiraju (vbhupati@#brocade.com)
# Shiv Haris (sharis@brocade.com)


"""Brocade NOS Driver CLI."""
from __future__ import print_function

import argparse

from neutron.openstack.common import log as logging
from neutron.plugins.brocade.nos import nosdriver as nos

LOG = logging.getLogger(__name__)


class NOSCli(object):

    def __init__(self, host, username, password):
        self.host = host
        self.username = username
        self.password = password
        self.driver = nos.NOSdriver()

    def execute(self, cmd):
        numargs = len(args.otherargs)

        if args.cmd == 'create' and numargs == 1:
            self._create(args.otherargs[0])
        elif args.cmd == 'delete' and numargs == 1:
            self._delete(args.otherargs[0])
        elif args.cmd == 'associate' and numargs == 2:
            self._associate(args.otherargs[0], args.otherargs[1])
        elif args.cmd == 'dissociate' and numargs == 2:
            self._dissociate(args.otherargs[0], args.otherargs[1])
        else:
            print(usage_desc)
            exit(0)

    def _create(self, net_id):
        self.driver.create_network(self.host, self.username, self.password,
                                   net_id)

    def _delete(self, net_id):
        self.driver.delete_network(self.host, self.username, self.password,
                                   net_id)

    def _associate(self, net_id, mac):
        self.driver.associate_mac_to_network(
            self.host, self.username, self.password, net_id, mac)

    def _dissociate(self, net_id, mac):
        self.driver.dissociate_mac_from_network(
            self.host, self.username, self.password, net_id, mac)


usage_desc = """
Command descriptions:

    create <id>
    delete <id>
    associate <id> <mac>
    dissociate <id> <mac>
"""

parser = argparse.ArgumentParser(description='process args',
                                 usage=usage_desc, epilog='foo bar help')
parser.add_argument('--ip', default='localhost')
parser.add_argument('--username', default='admin')
parser.add_argument('--password', default='password')
parser.add_argument('cmd')
parser.add_argument('otherargs', nargs='*')
args = parser.parse_args()

noscli = NOSCli(args.ip, args.username, args.password)
noscli.execute(args.cmd)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""LBaaS add status description

Revision ID: 2032abe8edac
Revises: 477a4488d3f4
Create Date: 2013-06-24 06:51:47.308545

"""

# revision identifiers, used by Alembic.
revision = '2032abe8edac'
down_revision = '477a4488d3f4'

# Change to ['*'] if this migration applies to all plugins

migration_for_plugins = [
    'neutron.services.loadbalancer.plugin.LoadBalancerPlugin',
]

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

ENTITIES = ['vips', 'pools', 'members', 'healthmonitors']


def upgrade(active_plugins=None, options=None):
    if not migration.should_run(active_plugins, migration_for_plugins):
        return

    for entity in ENTITIES:
        op.add_column(entity, sa.Column('status_description', sa.String(255)))


def downgrade(active_plugins=None, options=None):
    if not migration.should_run(active_plugins, migration_for_plugins):
        return

    for entity in ENTITIES:
        op.drop_column(entity, 'status_description')



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo.config import cfg

from neutron.agent.common import config as agent_config
from neutron.agent import l3_agent
from neutron.agent.linux import interface
from neutron.common import config as base_config
from neutron.openstack.common import uuidutils
from neutron.services.vpn import agent
from neutron.services.vpn import device_drivers
from neutron.tests import base

_uuid = uuidutils.generate_uuid
NOOP_DEVICE_CLASS = 'NoopDeviceDriver'
NOOP_DEVICE = ('neutron.tests.unit.services.'
               'vpn.test_vpn_agent.%s' % NOOP_DEVICE_CLASS)


class NoopDeviceDriver(device_drivers.DeviceDriver):
    def sync(self, context, processes):
        pass

    def create_router(self, process_id):
        pass

    def destroy_router(self, process_id):
        pass


class TestVPNAgent(base.BaseTestCase):
    def setUp(self):
        super(TestVPNAgent, self).setUp()
        self.conf = cfg.CONF
        self.conf.register_opts(base_config.core_opts)
        self.conf.register_opts(l3_agent.L3NATAgent.OPTS)
        self.conf.register_opts(interface.OPTS)
        agent_config.register_interface_driver_opts_helper(self.conf)
        agent_config.register_use_namespaces_opts_helper(self.conf)
        agent_config.register_agent_state_opts_helper(self.conf)
        agent_config.register_root_helper(self.conf)

        self.conf.set_override('interface_driver',
                               'neutron.agent.linux.interface.NullDriver')
        self.conf.set_override(
            'vpn_device_driver',
            [NOOP_DEVICE],
            'vpnagent')

        for clazz in [
            'neutron.agent.linux.ip_lib.device_exists',
            'neutron.agent.linux.ip_lib.IPWrapper',
            'neutron.agent.linux.interface.NullDriver',
            'neutron.agent.linux.utils.execute'
        ]:
            mock.patch(clazz).start()

        l3pluginApi_cls = mock.patch(
            'neutron.agent.l3_agent.L3PluginApi').start()
        self.plugin_api = mock.Mock()
        l3pluginApi_cls.return_value = self.plugin_api

        looping_call_p = mock.patch(
            'neutron.openstack.common.loopingcall.FixedIntervalLoopingCall')
        looping_call_p.start()

        self.fake_host = 'fake_host'
        self.agent = agent.VPNAgent(self.fake_host)

    def test_setup_drivers(self):
        self.assertEqual(1, len(self.agent.devices))
        device = self.agent.devices[0]
        self.assertEqual(
            NOOP_DEVICE_CLASS,
            device.__class__.__name__
        )

    def test_get_namespace(self):
        router_id = _uuid()
        ri = l3_agent.RouterInfo(router_id, self.conf.root_helper,
                                 self.conf.use_namespaces, None)
        self.agent.router_info = {router_id: ri}
        namespace = self.agent.get_namespace(router_id)
        self.assertTrue(namespace.endswith(router_id))
        self.assertFalse(self.agent.get_namespace('fake_id'))

    def test_add_nat_rule(self):
        router_id = _uuid()
        ri = l3_agent.RouterInfo(router_id, self.conf.root_helper,
                                 self.conf.use_namespaces, None)
        iptables = mock.Mock()
        ri.iptables_manager.ipv4['nat'] = iptables
        self.agent.router_info = {router_id: ri}
        self.agent.add_nat_rule(router_id, 'fake_chain', 'fake_rule', True)
        iptables.add_rule.assert_called_once_with(
            'fake_chain', 'fake_rule', top=True)

    def test_add_nat_rule_with_no_router(self):
        self.agent.router_info = {}
        #Should do nothing
        self.agent.add_nat_rule(
            'fake_router_id',
            'fake_chain',
            'fake_rule',
            True)

    def test_remove_rule(self):
        router_id = _uuid()
        ri = l3_agent.RouterInfo(router_id, self.conf.root_helper,
                                 self.conf.use_namespaces, None)
        iptables = mock.Mock()
        ri.iptables_manager.ipv4['nat'] = iptables
        self.agent.router_info = {router_id: ri}
        self.agent.remove_nat_rule(router_id, 'fake_chain', 'fake_rule')
        iptables.remove_rule.assert_called_once_with(
            'fake_chain', 'fake_rule')

    def test_remove_rule_with_no_router(self):
        self.agent.router_info = {}
        #Should do nothing
        self.agent.remove_nat_rule(
            'fake_router_id',
            'fake_chain',
            'fake_rule')

    def test_iptables_apply(self):
        router_id = _uuid()
        ri = l3_agent.RouterInfo(router_id, self.conf.root_helper,
                                 self.conf.use_namespaces, None)
        iptables = mock.Mock()
        ri.iptables_manager = iptables
        self.agent.router_info = {router_id: ri}
        self.agent.iptables_apply(router_id)
        iptables.apply.assert_called_once_with()

    def test_iptables_apply_with_no_router(self):
        #Should do nothing
        self.agent.router_info = {}
        self.agent.iptables_apply('fake_router_id')

    def test_router_added(self):
        mock.patch(
            'neutron.agent.linux.iptables_manager.IptablesManager').start()
        router_id = _uuid()
        router = {'id': router_id}
        device = mock.Mock()
        self.agent.devices = [device]
        self.agent._router_added(router_id, router)
        device.create_router.assert_called_once_with(router_id)

    def test_router_removed(self):
        self.plugin_api.get_external_network_id.return_value = None
        mock.patch(
            'neutron.agent.linux.iptables_manager.IptablesManager').start()
        router_id = _uuid()
        ri = l3_agent.RouterInfo(router_id, self.conf.root_helper,
                                 self.conf.use_namespaces, None)
        ri.router = {
            'id': _uuid(),
            'admin_state_up': True,
            'routes': [],
            'external_gateway_info': {}}
        device = mock.Mock()
        self.agent.router_info = {router_id: ri}
        self.agent.devices = [device]
        self.agent._router_removed(router_id)
        device.destroy_router.assert_called_once_with(router_id)

    def test_process_routers(self):
        self.plugin_api.get_external_network_id.return_value = None
        routers = [
            {'id': _uuid(),
             'admin_state_up': True,
             'routes': [],
             'external_gateway_info': {}}]

        device = mock.Mock()
        self.agent.devices = [device]
        self.agent._process_routers(routers, False)
        device.sync.assert_called_once_with(mock.ANY, routers)



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc

from neutron.api.v2 import attributes
from neutron.db import db_base_plugin_v2
from neutron.db import model_base
from neutron.db import models_v2
from neutron.openstack.common import log as logging
from neutron.plugins.vmware.extensions import maclearning as mac

LOG = logging.getLogger(__name__)


class MacLearningState(model_base.BASEV2):

    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id', ondelete="CASCADE"),
                        primary_key=True)
    mac_learning_enabled = sa.Column(sa.Boolean(), nullable=False)

    # Add a relationship to the Port model using the backref attribute.
    # This will instruct SQLAlchemy to eagerly load this association.
    port = orm.relationship(
        models_v2.Port,
        backref=orm.backref("mac_learning_state", lazy='joined',
                            uselist=False, cascade='delete'))


class MacLearningDbMixin(object):
    """Mixin class for mac learning."""

    def _make_mac_learning_state_dict(self, port, fields=None):
        res = {'port_id': port['port_id'],
               mac.MAC_LEARNING: port[mac.MAC_LEARNING]}
        return self._fields(res, fields)

    def _extend_port_mac_learning_state(self, port_res, port_db):
        state = port_db.mac_learning_state
        if state and state.mac_learning_enabled:
            port_res[mac.MAC_LEARNING] = state.mac_learning_enabled

    # Register dict extend functions for ports
    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
        attributes.PORTS, ['_extend_port_mac_learning_state'])

    def _update_mac_learning_state(self, context, port_id, enabled):
        try:
            query = self._model_query(context, MacLearningState)
            state = query.filter(MacLearningState.port_id == port_id).one()
            state.update({mac.MAC_LEARNING: enabled})
        except exc.NoResultFound:
            self._create_mac_learning_state(context,
                                            {'id': port_id,
                                             mac.MAC_LEARNING: enabled})

    def _create_mac_learning_state(self, context, port):
        with context.session.begin(subtransactions=True):
            enabled = port[mac.MAC_LEARNING]
            state = MacLearningState(port_id=port['id'],
                                     mac_learning_enabled=enabled)
            context.session.add(state)
        return self._make_mac_learning_state_dict(state)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import mock
import testtools

from neutron.db import api as db
from neutron.plugins.cisco.common import cisco_exceptions as c_exc
from neutron.plugins.cisco.common import config
from neutron.plugins.cisco.db import nexus_db_v2 as nxdb
from neutron.plugins.cisco.nexus import cisco_nexus_plugin_v2
from neutron.tests import base


class CiscoNexusDbTest(base.BaseTestCase):

    """Unit tests for cisco.db.nexus_models_v2.NexusPortBinding model."""

    NpbObj = collections.namedtuple('NpbObj', 'port vlan switch instance')

    def setUp(self):
        super(CiscoNexusDbTest, self).setUp()
        db.configure_db()
        self.session = db.get_session()
        self.addCleanup(db.clear_db)

    def _npb_test_obj(self, pnum, vnum, switch=None, instance=None):
        """Create a Nexus port binding test object from a pair of numbers."""
        if pnum is 'router':
            port = pnum
        else:
            port = '1/%s' % str(pnum)
        vlan = str(vnum)
        if switch is None:
            switch = '10.9.8.7'
        if instance is None:
            instance = 'instance_%s_%s' % (str(pnum), str(vnum))
        return self.NpbObj(port, vlan, switch, instance)

    def _assert_equal(self, npb, npb_obj):
        self.assertEqual(npb.port_id, npb_obj.port)
        self.assertEqual(int(npb.vlan_id), int(npb_obj.vlan))
        self.assertEqual(npb.switch_ip, npb_obj.switch)
        self.assertEqual(npb.instance_id, npb_obj.instance)

    def _add_to_db(self, npbs):
        for npb in npbs:
            nxdb.add_nexusport_binding(
                npb.port, npb.vlan, npb.switch, npb.instance)

    def test_nexusportbinding_add_remove(self):
        npb11 = self._npb_test_obj(10, 100)
        npb = nxdb.add_nexusport_binding(
            npb11.port, npb11.vlan, npb11.switch, npb11.instance)
        self._assert_equal(npb, npb11)
        npb = nxdb.remove_nexusport_binding(
            npb11.port, npb11.vlan, npb11.switch, npb11.instance)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb11)
        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.remove_nexusport_binding(
                npb11.port, npb11.vlan, npb11.switch, npb11.instance)

    def test_nexusportbinding_get(self):
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_to_db([npb11, npb21, npb22])

        npb = nxdb.get_nexusport_binding(
            npb11.port, npb11.vlan, npb11.switch, npb11.instance)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb11)
        npb = nxdb.get_nexusport_binding(
            npb21.port, npb21.vlan, npb21.switch, npb21.instance)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb21)
        npb = nxdb.get_nexusport_binding(
            npb22.port, npb22.vlan, npb22.switch, npb22.instance)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb22)

        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.get_nexusport_binding(
                npb21.port, npb21.vlan, npb21.switch, "dummyInstance")

    def test_nexusvlanbinding_get(self):
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_to_db([npb11, npb21, npb22])

        npb_all_v100 = nxdb.get_nexusvlan_binding(npb11.vlan, npb11.switch)
        self.assertEqual(len(npb_all_v100), 2)
        npb_v200 = nxdb.get_nexusvlan_binding(npb22.vlan, npb22.switch)
        self.assertEqual(len(npb_v200), 1)
        self._assert_equal(npb_v200[0], npb22)

        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.get_nexusvlan_binding(npb21.vlan, "dummySwitch")

    def test_nexusvmbinding_get(self):
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_to_db([npb11, npb21, npb22])

        npb = nxdb.get_nexusvm_bindings(npb21.vlan, npb21.instance)[0]
        self._assert_equal(npb, npb21)
        npb = nxdb.get_nexusvm_bindings(npb22.vlan, npb22.instance)[0]
        self._assert_equal(npb, npb22)

        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.get_nexusvm_bindings(npb21.vlan, "dummyInstance")

    def test_nexusportvlanswitchbinding_get(self):
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        self._add_to_db([npb11, npb21])

        npb = nxdb.get_port_vlan_switch_binding(
            npb11.port, npb11.vlan, npb11.switch)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb11)

        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.get_port_vlan_switch_binding(
                npb21.port, npb21.vlan, "dummySwitch")

    def test_nexusportswitchbinding_get(self):
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100, switch='2.2.2.2')
        npb22 = self._npb_test_obj(20, 200, switch='2.2.2.2')
        self._add_to_db([npb11, npb21, npb22])

        npb = nxdb.get_port_switch_bindings(npb11.port, npb11.switch)
        self.assertEqual(len(npb), 1)
        self._assert_equal(npb[0], npb11)
        npb_all_p20 = nxdb.get_port_switch_bindings(npb21.port, npb21.switch)
        self.assertEqual(len(npb_all_p20), 2)

        npb = nxdb.get_port_switch_bindings(npb21.port, "dummySwitch")
        self.assertIsNone(npb)

    def test_nexussvibinding_get(self):
        npbr1 = self._npb_test_obj('router', 100)
        npb21 = self._npb_test_obj(20, 100)
        self._add_to_db([npbr1, npb21])

        npb_svi = nxdb.get_nexussvi_bindings()
        self.assertEqual(len(npb_svi), 1)
        self._assert_equal(npb_svi[0], npbr1)

        npbr2 = self._npb_test_obj('router', 200)
        self._add_to_db([npbr2])
        npb_svi = nxdb.get_nexussvi_bindings()
        self.assertEqual(len(npb_svi), 2)

    def test_nexussviswitch_find(self):
        """Test Nexus switch selection for SVI placement."""
        # Configure 2 Nexus switches
        nexus_switches = {
            ('1.1.1.1', 'username'): 'admin',
            ('1.1.1.1', 'password'): 'password1',
            ('1.1.1.1', 'host1'): '1/1',
            ('2.2.2.2', 'username'): 'admin',
            ('2.2.2.2', 'password'): 'password2',
            ('2.2.2.2', 'host2'): '1/1',
        }
        nexus_plugin = cisco_nexus_plugin_v2.NexusPlugin()
        nexus_plugin._client = mock.Mock()
        nexus_plugin._client.nexus_switches = nexus_switches

        # Set the Cisco config module's first configured device IP address
        # according to the preceding switch config
        with mock.patch.object(config, 'first_device_ip', new='1.1.1.1'):

            # Enable round-robin mode with no SVIs configured on any of the
            # Nexus switches (i.e. no entries in the SVI database). The
            # plugin should select the first switch in the configuration.
            config.CONF.set_override('svi_round_robin', True, 'CISCO')
            switch_ip = nexus_plugin._find_switch_for_svi()
            self.assertEqual(switch_ip, '1.1.1.1')

            # Keep round-robin mode enabled, and add entries to the SVI
            # database. The plugin should select the switch with the least
            # number of entries in the SVI database.
            vlan = 100
            npbr11 = self._npb_test_obj('router', vlan, switch='1.1.1.1',
                                        instance='instance11')
            npbr12 = self._npb_test_obj('router', vlan, switch='1.1.1.1',
                                        instance='instance12')
            npbr21 = self._npb_test_obj('router', vlan, switch='2.2.2.2',
                                        instance='instance21')
            self._add_to_db([npbr11, npbr12, npbr21])
            switch_ip = nexus_plugin._find_switch_for_svi()
            self.assertEqual(switch_ip, '2.2.2.2')

            # Disable round-robin mode. The plugin should select the
            # first switch in the configuration.
            config.CONF.clear_override('svi_round_robin', 'CISCO')
            switch_ip = nexus_plugin._find_switch_for_svi()
            self.assertEqual(switch_ip, '1.1.1.1')

    def test_nexusbinding_update(self):
        npb11 = self._npb_test_obj(10, 100, switch='1.1.1.1', instance='test')
        npb21 = self._npb_test_obj(20, 100, switch='1.1.1.1', instance='test')
        self._add_to_db([npb11, npb21])

        npb_all_v100 = nxdb.get_nexusvlan_binding(npb11.vlan, '1.1.1.1')
        self.assertEqual(len(npb_all_v100), 2)

        npb22 = self._npb_test_obj(20, 200, switch='1.1.1.1', instance='test')
        npb = nxdb.update_nexusport_binding(npb21.port, 200)
        self._assert_equal(npb, npb22)

        npb_all_v100 = nxdb.get_nexusvlan_binding(npb11.vlan, '1.1.1.1')
        self.assertEqual(len(npb_all_v100), 1)
        self._assert_equal(npb_all_v100[0], npb11)

        npb = nxdb.update_nexusport_binding(npb21.port, 0)
        self.assertIsNone(npb)

        npb33 = self._npb_test_obj(30, 300, switch='1.1.1.1', instance='test')
        with testtools.ExpectedException(c_exc.NexusPortBindingNotFound):
            nxdb.update_nexusport_binding(npb33.port, 200)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# @author: Ryota MIBU

import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db import model_base
from neutron.db import models_v2


"""New mapping tables."""


class OFCId(object):
    """Resource ID on OpenFlow Controller."""
    ofc_id = sa.Column(sa.String(255), unique=True, nullable=False)


class NeutronId(object):
    """Logical ID on Neutron."""
    neutron_id = sa.Column(sa.String(36), primary_key=True)


class OFCTenantMapping(model_base.BASEV2, NeutronId, OFCId):
    """Represents a Tenant on OpenFlow Network/Controller."""


class OFCNetworkMapping(model_base.BASEV2, NeutronId, OFCId):
    """Represents a Network on OpenFlow Network/Controller."""


class OFCPortMapping(model_base.BASEV2, NeutronId, OFCId):
    """Represents a Port on OpenFlow Network/Controller."""


class OFCRouterMapping(model_base.BASEV2, NeutronId, OFCId):
    """Represents a router on OpenFlow Network/Controller."""


class OFCFilterMapping(model_base.BASEV2, NeutronId, OFCId):
    """Represents a Filter on OpenFlow Network/Controller."""


class PortInfo(model_base.BASEV2):
    """Represents a Virtual Interface."""
    id = sa.Column(sa.String(36),
                   sa.ForeignKey('ports.id', ondelete="CASCADE"),
                   primary_key=True)
    datapath_id = sa.Column(sa.String(36), nullable=False)
    port_no = sa.Column(sa.Integer, nullable=False)
    vlan_id = sa.Column(sa.Integer, nullable=False)
    mac = sa.Column(sa.String(32), nullable=False)
    port = orm.relationship(
        models_v2.Port,
        backref=orm.backref("portinfo",
                            lazy='joined', uselist=False,
                            cascade='delete'))



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2012 New Dream Network, LLC (DreamHost)
# Author: Julien Danjou <julien@danjou.info>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Cron script to generate usage notifications for networks, ports and
subnets.

"""

from oslo.config import cfg

from neutron.common import config
from neutron import context
from neutron import manager
from neutron.openstack.common.notifier import api as notifier_api


def main():
    cfg.CONF(project='neutron')
    config.setup_logging(cfg.CONF)

    cxt = context.get_admin_context()
    plugin = manager.NeutronManager.get_plugin()
    for network in plugin.get_networks(cxt):
        notifier_api.notify(cxt,
                            notifier_api.publisher_id('network'),
                            'network.exists',
                            notifier_api.INFO,
                            {'network': network})
    for subnet in plugin.get_subnets(cxt):
        notifier_api.notify(cxt,
                            notifier_api.publisher_id('network'),
                            'subnet.exists',
                            notifier_api.INFO,
                            {'subnet': subnet})
    for port in plugin.get_ports(cxt):
        notifier_api.notify(cxt,
                            notifier_api.publisher_id('network'),
                            'port.exists',
                            notifier_api.INFO,
                            {'port': port})
    for router in plugin.get_routers(cxt):
        notifier_api.notify(cxt,
                            notifier_api.publisher_id('network'),
                            'router.exists',
                            notifier_api.INFO,
                            {'router': router})
    for floatingip in plugin.get_floatingips(cxt):
        notifier_api.notify(cxt,
                            notifier_api.publisher_id('network'),
                            'floatingip.exists',
                            notifier_api.INFO,
                            {'floatingip': floatingip})



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Sumit Naiksatam, Cisco Systems, Inc.
# @author: Rohit Agarwalla, Cisco Systems, Inc.
#

import inspect
import logging
import sys

from neutron.api.v2 import attributes
from neutron.db import api as db_api
from neutron.extensions import portbindings
from neutron.extensions import providernet as provider
from neutron import neutron_plugin_base_v2
from neutron.openstack.common import importutils
from neutron.plugins.cisco.common import cisco_constants as const
from neutron.plugins.cisco.common import cisco_credentials_v2 as cred
from neutron.plugins.cisco.common import cisco_exceptions as cexc
from neutron.plugins.cisco.common import config as conf
from neutron.plugins.cisco.db import network_db_v2 as cdb
from neutron.plugins.openvswitch import ovs_db_v2 as odb


LOG = logging.getLogger(__name__)


class VirtualPhysicalSwitchModelV2(neutron_plugin_base_v2.NeutronPluginBaseV2):
    """Virtual Physical Switch Model.

    This implementation works with OVS and Nexus plugin for the
    following topology:
    One or more servers to a nexus switch.
    """
    __native_bulk_support = True
    supported_extension_aliases = ["provider", "binding"]
    _methods_to_delegate = ['create_network_bulk',
                            'get_network', 'get_networks',
                            'create_port_bulk',
                            'get_port', 'get_ports',
                            'create_subnet', 'create_subnet_bulk',
                            'delete_subnet', 'update_subnet',
                            'get_subnet', 'get_subnets',
                            'create_or_update_agent', 'report_state']

    def __init__(self):
        """Initialize the segmentation manager.

        Checks which device plugins are configured, and load the inventories
        those device plugins for which the inventory is configured.
        """
        conf.CiscoConfigOptions()

        self._plugins = {}
        for key in conf.CISCO_PLUGINS.keys():
            plugin_obj = conf.CISCO_PLUGINS[key]
            if plugin_obj is not None:
                self._plugins[key] = importutils.import_object(plugin_obj)
                LOG.debug(_("Loaded device plugin %s"),
                          conf.CISCO_PLUGINS[key])

        if ((const.VSWITCH_PLUGIN in self._plugins) and
            hasattr(self._plugins[const.VSWITCH_PLUGIN],
                    "supported_extension_aliases")):
            self.supported_extension_aliases.extend(
                self._plugins[const.VSWITCH_PLUGIN].
                supported_extension_aliases)
        # At this point, all the database models should have been loaded. It's
        # possible that configure_db() may have been called by one of the
        # plugins loaded in above. Otherwise, this call is to make sure that
        # the database is initialized
        db_api.configure_db()

        # Initialize credential store after database initialization
        cred.Store.initialize()
        LOG.debug(_("%(module)s.%(name)s init done"),
                  {'module': __name__,
                   'name': self.__class__.__name__})

        # Check whether we have a valid Nexus driver loaded
        self.is_nexus_plugin = False
        nexus_driver = conf.CISCO.nexus_driver
        if nexus_driver.endswith('CiscoNEXUSDriver'):
            self.is_nexus_plugin = True

    def __getattribute__(self, name):
        """Delegate calls to OVS sub-plugin.

        This delegates the calls to the methods implemented only by the OVS
        sub-plugin. Note: Currently, bulking is handled by the caller
        (PluginV2), and this model class expects to receive only non-bulking
        calls. If, however, a bulking call is made, this will method will
        delegate the call to the OVS plugin.
        """
        super_getattribute = super(VirtualPhysicalSwitchModelV2,
                                   self).__getattribute__
        methods = super_getattribute('_methods_to_delegate')

        if name in methods:
            plugin = super_getattribute('_plugins')[const.VSWITCH_PLUGIN]
            return getattr(plugin, name)

        try:
            return super_getattribute(name)
        except AttributeError:
            plugin = super_getattribute('_plugins')[const.VSWITCH_PLUGIN]
            return getattr(plugin, name)

    def _func_name(self, offset=0):
        """Get the name of the calling function."""
        frame_record = inspect.stack()[1 + offset]
        func_name = frame_record[3]
        return func_name

    def _invoke_plugin_per_device(self, plugin_key, function_name,
                                  args, **kwargs):
        """Invoke plugin per device.

        Invokes a device plugin's relevant functions (based on the
        plugin implementation) for completing this operation.
        """
        if plugin_key not in self._plugins:
            LOG.info(_("No %s Plugin loaded"), plugin_key)
            LOG.info(_("%(plugin_key)s: %(function_name)s with args %(args)s "
                     "ignored"),
                     {'plugin_key': plugin_key,
                      'function_name': function_name,
                      'args': args})
        else:
            func = getattr(self._plugins[plugin_key], function_name)
            return func(*args, **kwargs)

    def _get_segmentation_id(self, network_id):
        binding_seg_id = odb.get_network_binding(None, network_id)
        if not binding_seg_id:
            raise cexc.NetworkSegmentIDNotFound(net_id=network_id)
        return binding_seg_id.segmentation_id

    def _get_provider_vlan_id(self, network):
        if (all(attributes.is_attr_set(network.get(attr))
                for attr in (provider.NETWORK_TYPE,
                             provider.PHYSICAL_NETWORK,
                             provider.SEGMENTATION_ID))
            and
                network[provider.NETWORK_TYPE] == const.NETWORK_TYPE_VLAN):
            return network[provider.SEGMENTATION_ID]

    def create_network(self, context, network):
        """Create network.

        Perform this operation in the context of the configured device
        plugins.
        """
        LOG.debug(_("create_network() called"))
        provider_vlan_id = self._get_provider_vlan_id(network[const.NETWORK])
        args = [context, network]
        ovs_output = self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                    self._func_name(),
                                                    args)
        # The vswitch plugin did all the verification. If it's a provider
        # vlan network, save it for the nexus plugin to use later.
        if provider_vlan_id:
            network_id = ovs_output[const.NET_ID]
            cdb.add_provider_network(network_id,
                                     const.NETWORK_TYPE_VLAN,
                                     provider_vlan_id)
            LOG.debug(_("Provider network added to DB: %(network_id)s, "
                        "%(vlan_id)s"),
                      {'network_id': network_id, 'vlan_id': provider_vlan_id})
        return ovs_output

    def update_network(self, context, id, network):
        """Update network.

        Perform this operation in the context of the configured device
        plugins.

        Note that the Nexus sub-plugin does not need to be notified
        (and the Nexus switch does not need to be [re]configured)
        for an update network operation because the Nexus sub-plugin
        is agnostic of all network-level attributes except the
        segmentation ID. Furthermore, updating of the segmentation ID
        is not supported by the OVS plugin since it is considered a
        provider attribute, so it is not supported by this method.
        """
        LOG.debug(_("update_network() called"))

        # We can only support updating of provider attributes if all the
        # configured sub-plugins support it. Currently we have no method
        # in place for checking whether a sub-plugin supports it,
        # so assume not.
        provider._raise_if_updates_provider_attributes(network['network'])

        args = [context, id, network]
        return self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                              self._func_name(),
                                              args)

    def delete_network(self, context, id):
        """Delete network.

        Perform this operation in the context of the configured device
        plugins.
        """
        args = [context, id]
        ovs_output = self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                    self._func_name(),
                                                    args)
        if cdb.remove_provider_network(id):
            LOG.debug(_("Provider network removed from DB: %s"), id)
        return ovs_output

    def get_network(self, context, id, fields=None):
        """Get network. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def get_networks(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None, page_reverse=False):
        """Get networks. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def _invoke_nexus_for_net_create(self, context, tenant_id, net_id,
                                     instance_id, host_id):
        if not self.is_nexus_plugin:
            return False

        network = self.get_network(context, net_id)
        vlan_id = self._get_segmentation_id(net_id)
        vlan_name = conf.CISCO.vlan_name_prefix + str(vlan_id)
        network[const.NET_VLAN_ID] = vlan_id
        network[const.NET_VLAN_NAME] = vlan_name
        attachment = {
            const.TENANT_ID: tenant_id,
            const.INSTANCE_ID: instance_id,
            const.HOST_NAME: host_id,
        }
        self._invoke_plugin_per_device(
            const.NEXUS_PLUGIN,
            'create_network',
            [network, attachment])

    def _check_valid_port_device_owner(self, port):
        """Check the port for valid device_owner.

        Don't call the nexus plugin for router and dhcp
        port owners.
        """
        return port['device_owner'].startswith('compute')

    def _get_port_host_id_from_bindings(self, port):
        """Get host_id from portbindings."""
        host_id = None

        if (portbindings.HOST_ID in port and
            attributes.is_attr_set(port[portbindings.HOST_ID])):
            host_id = port[portbindings.HOST_ID]

        return host_id

    def create_port(self, context, port):
        """Create port.

        Perform this operation in the context of the configured device
        plugins.
        """
        LOG.debug(_("create_port() called"))
        args = [context, port]
        ovs_output = self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                    self._func_name(),
                                                    args)
        instance_id = port['port']['device_id']

        # Only call nexus plugin if there's a valid instance_id, host_id
        # and device_owner
        try:
            host_id = self._get_port_host_id_from_bindings(port['port'])
            if (instance_id and host_id and
                self._check_valid_port_device_owner(port['port'])):
                net_id = port['port']['network_id']
                tenant_id = port['port']['tenant_id']
                self._invoke_nexus_for_net_create(
                    context, tenant_id, net_id, instance_id, host_id)
        except Exception:
            # Create network on the Nexus plugin has failed, so we need
            # to rollback the port creation on the VSwitch plugin.
            exc_info = sys.exc_info()
            try:
                id = ovs_output['id']
                args = [context, id]
                ovs_output = self._invoke_plugin_per_device(
                    const.VSWITCH_PLUGIN,
                    'delete_port',
                    args)
            finally:
                # Re-raise the original exception
                raise exc_info[0], exc_info[1], exc_info[2]
        return ovs_output

    def get_port(self, context, id, fields=None):
        """Get port. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def get_ports(self, context, filters=None, fields=None):
        """Get ports. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def _check_nexus_net_create_needed(self, new_port, old_port):
        """Check if nexus plugin should be invoked for net_create.

        In the following cases, the plugin should be invoked:
           -- a port is attached to a VM instance. The old host id is None
           -- VM migration. The old host id has a valid value

        When the plugin needs to be invoked, return the old_host_id,
        and a list of calling arguments.
        Otherwise, return '' for old host id and an empty list
        """
        old_device_id = old_port['device_id']
        new_device_id = new_port.get('device_id')
        new_host_id = self._get_port_host_id_from_bindings(new_port)
        tenant_id = old_port['tenant_id']
        net_id = old_port['network_id']
        old_host_id = self._get_port_host_id_from_bindings(old_port)

        LOG.debug(_("tenant_id: %(tid)s, net_id: %(nid)s, "
                    "old_device_id: %(odi)s, new_device_id: %(ndi)s, "
                    "old_host_id: %(ohi)s, new_host_id: %(nhi)s, "
                    "old_device_owner: %(odo)s, new_device_owner: %(ndo)s"),
                  {'tid': tenant_id, 'nid': net_id,
                   'odi': old_device_id, 'ndi': new_device_id,
                   'ohi': old_host_id, 'nhi': new_host_id,
                   'odo': old_port.get('device_owner'),
                   'ndo': new_port.get('device_owner')})

        # A port is attached to an instance
        if (new_device_id and not old_device_id and new_host_id and
                self._check_valid_port_device_owner(new_port)):
            return '', [tenant_id, net_id, new_device_id, new_host_id]

        # An instance is being migrated
        if (old_device_id and old_host_id and
                new_host_id and new_host_id != old_host_id and
                self._check_valid_port_device_owner(old_port)):
            return old_host_id, [tenant_id, net_id, old_device_id, new_host_id]

        # no need to invoke the plugin
        return '', []

    def update_port(self, context, id, port):
        """Update port.

        Perform this operation in the context of the configured device
        plugins.
        """
        LOG.debug(_("update_port() called"))
        old_port = self.get_port(context, id)
        args = [context, id, port]
        ovs_output = self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                    self._func_name(),
                                                    args)
        try:
            # Check if the nexus plugin needs to be invoked
            if self.is_nexus_plugin:
                old_host_id, create_args = self._check_nexus_net_create_needed(
                    port['port'], old_port)

                # In the case of migration, invoke it to remove
                # the previous port binding
                if old_host_id:
                    vlan_id = self._get_segmentation_id(old_port['network_id'])
                    delete_args = [old_port['device_id'], vlan_id]
                    self._invoke_plugin_per_device(const.NEXUS_PLUGIN,
                                                   "delete_port",
                                                   delete_args)

                # Invoke the Nexus plugin to create a net and/or new
                # port binding
                if create_args:
                    self._invoke_nexus_for_net_create(context, *create_args)

            return ovs_output
        except Exception:
            exc_info = sys.exc_info()
            LOG.error(_("Unable to update port '%s' on Nexus switch"),
                      old_port['name'], exc_info=exc_info)
            try:
                # Roll back vSwitch plugin to original port attributes.
                args = [context, id, {'port': old_port}]
                self._invoke_plugin_per_device(
                    const.VSWITCH_PLUGIN,
                    self._func_name(),
                    args)
            finally:
                # Re-raise the original exception
                raise exc_info[0], exc_info[1], exc_info[2]

    def delete_port(self, context, id, l3_port_check=True):
        """Delete port.

        Perform this operation in the context of the configured device
        plugins.
        """
        LOG.debug(_("delete_port() called"))
        port = self.get_port(context, id)

        host_id = self._get_port_host_id_from_bindings(port)

        if (self.is_nexus_plugin and host_id and
            self._check_valid_port_device_owner(port)):
            vlan_id = self._get_segmentation_id(port['network_id'])
            n_args = [port['device_id'], vlan_id]
            self._invoke_plugin_per_device(const.NEXUS_PLUGIN,
                                           self._func_name(),
                                           n_args)
        try:
            args = [context, id]
            ovs_output = self._invoke_plugin_per_device(
                const.VSWITCH_PLUGIN, self._func_name(),
                args, l3_port_check=l3_port_check)
        except Exception:
            exc_info = sys.exc_info()
            # Roll back the delete port on the Nexus plugin
            try:
                tenant_id = port['tenant_id']
                net_id = port['network_id']
                instance_id = port['device_id']
                host_id = port[portbindings.HOST_ID]
                self._invoke_nexus_for_net_create(context, tenant_id, net_id,
                                                  instance_id, host_id)
            finally:
                # Raise the original exception.
                raise exc_info[0], exc_info[1], exc_info[2]

        return ovs_output

    def add_router_interface(self, context, router_id, interface_info):
        """Add a router interface on a subnet.

        Only invoke the Nexus plugin to create SVI if L3 support on
        the Nexus switches is enabled and a Nexus plugin is loaded,
        otherwise send it to the vswitch plugin
        """
        if (conf.CISCO.nexus_l3_enable and self.is_nexus_plugin):
            LOG.debug(_("L3 enabled on Nexus plugin, create SVI on switch"))
            if 'subnet_id' not in interface_info:
                raise cexc.SubnetNotSpecified()
            if 'port_id' in interface_info:
                raise cexc.PortIdForNexusSvi()
            subnet = self.get_subnet(context, interface_info['subnet_id'])
            gateway_ip = subnet['gateway_ip']
            # Get gateway IP address and netmask
            cidr = subnet['cidr']
            netmask = cidr.split('/', 1)[1]
            gateway_ip = gateway_ip + '/' + netmask
            network_id = subnet['network_id']
            vlan_id = self._get_segmentation_id(network_id)
            vlan_name = conf.CISCO.vlan_name_prefix + str(vlan_id)

            n_args = [vlan_name, vlan_id, subnet['id'], gateway_ip, router_id]
            return self._invoke_plugin_per_device(const.NEXUS_PLUGIN,
                                                  self._func_name(),
                                                  n_args)
        else:
            LOG.debug(_("L3 disabled or not Nexus plugin, send to vswitch"))
            n_args = [context, router_id, interface_info]
            return self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                  self._func_name(),
                                                  n_args)

    def remove_router_interface(self, context, router_id, interface_info):
        """Remove a router interface.

        Only invoke the Nexus plugin to delete SVI if L3 support on
        the Nexus switches is enabled and a Nexus plugin is loaded,
        otherwise send it to the vswitch plugin
        """
        if (conf.CISCO.nexus_l3_enable and self.is_nexus_plugin):
            LOG.debug(_("L3 enabled on Nexus plugin, delete SVI from switch"))

            subnet = self.get_subnet(context, interface_info['subnet_id'])
            network_id = subnet['network_id']
            vlan_id = self._get_segmentation_id(network_id)
            n_args = [vlan_id, router_id]

            return self._invoke_plugin_per_device(const.NEXUS_PLUGIN,
                                                  self._func_name(),
                                                  n_args)
        else:
            LOG.debug(_("L3 disabled or not Nexus plugin, send to vswitch"))
            n_args = [context, router_id, interface_info]
            return self._invoke_plugin_per_device(const.VSWITCH_PLUGIN,
                                                  self._func_name(),
                                                  n_args)

    def create_subnet(self, context, subnet):
        """Create subnet. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def update_subnet(self, context, id, subnet):
        """Update subnet. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def get_subnet(self, context, id, fields=None):
        """Get subnet. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def delete_subnet(self, context, id, kwargs):
        """Delete subnet. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover

    def get_subnets(self, context, filters=None, fields=None,
                    sorts=None, limit=None, marker=None, page_reverse=False):
        """Get subnets. This method is delegated to the vswitch plugin.

        This method is included here to satisfy abstract method requirements.
        """
        pass  # pragma no cover
